# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional, Union

import torch
from torch import Tensor

from nncf.common.graph.graph import NNCFNode
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.graph.transformations.commands import TransformationPriority
from nncf.common.quantization.structs import NonWeightQuantizerId
from nncf.torch.graph.transformations.commands import ExtraCompressionModuleType
from nncf.torch.graph.transformations.commands import PTBiasCorrectionCommand
from nncf.torch.graph.transformations.commands import PTInsertionCommand
from nncf.torch.graph.transformations.commands import PTSharedFnInsertionCommand
from nncf.torch.graph.transformations.commands import PTTargetPoint
from nncf.torch.graph.transformations.commands import PTWeightUpdateCommand
from nncf.torch.quantization.layers import BaseQuantizer


def create_bias_correction_command(node: NNCFNode, bias_value: Tensor) -> PTBiasCorrectionCommand:
    """
     Creates bias correction command.

    :param node: The node in the NNCF graph that corresponds to operation with bias.
    :param bias_value: The new bias value that will be set.
    :return: The `PTBiasCorrectionCommand` command to update bias.
    """
    target_point = PTTargetPoint(TargetType.LAYER, node.node_name)
    return PTBiasCorrectionCommand(target_point, bias_value)


def create_command_to_update_weight(node: NNCFNode, weight_value: Tensor) -> PTWeightUpdateCommand:
    """
     Creates weight update command.

    :param node: The node in the NNCF graph that corresponds to operation with weight.
    :param weight_value: The new weight value that will be set.
    :return: The `PTWeightUpdateCommand` command to update weight.
    """
    target_point = PTTargetPoint(TargetType.LAYER, node.node_name)
    return PTWeightUpdateCommand(target_point, weight_value)


def create_quantizer_insertion_command(
    target_point: PTTargetPoint, quantizer: BaseQuantizer
) -> Union[PTInsertionCommand, PTSharedFnInsertionCommand]:
    quantizer_id = NonWeightQuantizerId(target_point.target_node_name, target_point.input_port_id)
    storage_key = str(quantizer_id)
    return PTSharedFnInsertionCommand(
        target_points=[target_point],
        fn=quantizer,
        op_unique_name=storage_key,
        compression_module_type=ExtraCompressionModuleType.EXTERNAL_QUANTIZER,
        priority=TransformationPriority.QUANTIZATION_PRIORITY,
    )


def create_shared_quantizer_insertion_command(
    target_points: List[PTTargetPoint], quantizer: BaseQuantizer
) -> PTSharedFnInsertionCommand:
    quantizers_ids = []
    for target_point in target_points:
        quantizers_ids.append(NonWeightQuantizerId(target_point.target_node_name, target_point.input_port_id))

    storage_key = ";".join(str(quantizer_id) for quantizer_id in sorted(quantizers_ids, key=str))
    return PTSharedFnInsertionCommand(
        target_points=target_points,
        fn=quantizer,
        op_unique_name=storage_key,
        compression_module_type=ExtraCompressionModuleType.EXTERNAL_QUANTIZER,
        priority=TransformationPriority.QUANTIZATION_PRIORITY,
    )


def create_pt_insertion_command(
    module: torch.nn.Module,
    target_type: TargetType,
    target_node_name: str,
    priority: int,
    input_port_id: Optional[int],
) -> PTInsertionCommand:
    """
    Creates a PTInsertionCommand.

    :param module: Torch module to insert.
    :param target_type: Insertion command target type.
    :param target_name: Insertion command target name.
    :param priority: Insertion command priority.
    :param input_port_id: Insertion command input port id.
    :return: A PTInsertionCommand
    """
    target_point = PTTargetPoint(
        target_type=target_type, target_node_name=target_node_name, input_port_id=input_port_id
    )
    return PTInsertionCommand(point=target_point, fn=module, priority=priority)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, List, TypeVar

import numpy as np

import nncf
from nncf.common.utils.backend import BackendType

TTensor = TypeVar("TTensor")


def create_normalized_mse_func(backend: BackendType) -> Callable[[List[TTensor], List[TTensor]], float]:
    """
    Factory method to create backend-specific implementation of the normalized_nmse.

    :param backend: A backend type.
    :return: The backend-specific implementation of the normalized_nmse.
    """
    if backend == BackendType.OPENVINO:
        return normalized_mse

    raise nncf.UnsupportedBackendError(
        f"Could not create backend-specific implementation! {backend} backend is not supported!"
    )


def normalized_mse(ref_outputs: List[np.ndarray], approx_outputs: List[np.ndarray]) -> float:
    """
    Calculates normalized mean square error between `ref_outputs` and `approx_outputs`.
    The normalized mean square error is defined as

    NMSE(x_ref, x_approx) = MSE(x_ref, x_approx) / MSE(x_ref, 0)

    :param ref_outputs: Reference outputs.
    :param approx_outputs: Approximate outputs.
    :return: The normalized mean square error between `ref_outputs` and `approx_outputs`.
    """
    metrics = []
    for x_ref, x_approx in zip(ref_outputs, approx_outputs):
        error_flattened = (x_ref - x_approx).flatten()
        x_ref_flattened = x_ref.flatten()
        nmse = np.dot(error_flattened, error_flattened) / np.dot(x_ref_flattened, x_ref_flattened)
        metrics.append(nmse)
    nmse = sum(metrics) / len(metrics)
    return nmse



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from copy import deepcopy
from typing import Set

import tensorflow as tf

from nncf import NNCFConfig
from nncf.api.compression import CompressionLoss
from nncf.api.compression import CompressionStage
from nncf.common.accuracy_aware_training.training_loop import ADAPTIVE_COMPRESSION_CONTROLLERS
from nncf.common.graph.operator_metatypes import OUTPUT_NOOP_METATYPES
from nncf.common.graph.transformations.commands import TransformationPriority
from nncf.common.initialization.batchnorm_adaptation import BatchnormAdaptationAlgorithm
from nncf.common.schedulers import StubCompressionScheduler
from nncf.common.scopes import check_scopes_in_graph
from nncf.common.scopes import should_consider_scope
from nncf.common.sparsity.schedulers import SPARSITY_SCHEDULERS
from nncf.common.sparsity.schedulers import SparsityScheduler
from nncf.common.sparsity.statistics import LayerThreshold
from nncf.common.sparsity.statistics import MagnitudeSparsityStatistics
from nncf.common.statistics import NNCFStatistics
from nncf.common.utils.api_marker import api
from nncf.config.extractors import extract_algo_specific_config
from nncf.config.extractors import extract_bn_adaptation_init_params
from nncf.config.schemata.defaults import MAGNITUDE_SPARSITY_WEIGHT_IMPORTANCE
from nncf.config.schemata.defaults import SPARSITY_INIT
from nncf.tensorflow.algorithm_selector import TF_COMPRESSION_ALGORITHMS
from nncf.tensorflow.api.compression import TFCompressionAlgorithmBuilder
from nncf.tensorflow.graph.converter import TFModelConverterFactory
from nncf.tensorflow.graph.metatypes.tf_ops import WEIGHTABLE_TF_OP_METATYPES
from nncf.tensorflow.graph.transformations.commands import TFInsertionCommand
from nncf.tensorflow.graph.transformations.commands import TFLayerWeight
from nncf.tensorflow.graph.transformations.layout import TFTransformationLayout
from nncf.tensorflow.graph.utils import collect_wrapped_layers
from nncf.tensorflow.graph.utils import get_original_name_and_instance_idx
from nncf.tensorflow.loss import TFZeroCompressionLoss
from nncf.tensorflow.sparsity.base_algorithm import SPARSITY_LAYER_METATYPES
from nncf.tensorflow.sparsity.base_algorithm import BaseSparsityController
from nncf.tensorflow.sparsity.collector import TFSparseModelStatisticsCollector
from nncf.tensorflow.sparsity.magnitude.functions import WEIGHT_IMPORTANCE_FUNCTIONS
from nncf.tensorflow.sparsity.magnitude.functions import calc_magnitude_binary_mask
from nncf.tensorflow.sparsity.magnitude.operation import BinaryMask
from nncf.tensorflow.sparsity.magnitude.operation import BinaryMaskWithWeightsBackup


@TF_COMPRESSION_ALGORITHMS.register("magnitude_sparsity")
class MagnitudeSparsityBuilder(TFCompressionAlgorithmBuilder):
    def __init__(self, config: NNCFConfig, should_init: bool = True):
        super().__init__(config, should_init)
        self.ignored_scopes = self._algo_config.get("ignored_scopes", [])
        self._op_names = []

    def get_transformation_layout(self, model: tf.keras.Model) -> TFTransformationLayout:
        converter = TFModelConverterFactory.create(model)
        nncf_graph = converter.convert()

        check_scopes_in_graph(nncf_graph, self.ignored_scopes, self.target_scopes, self.validate_scopes)

        transformations = TFTransformationLayout()

        processed_shared_layer_names: Set[str] = set()

        for node in nncf_graph.get_all_nodes():
            if node.is_shared():
                target_layer_name, _ = get_original_name_and_instance_idx(node.node_name)
                if target_layer_name in processed_shared_layer_names:
                    continue
                processed_shared_layer_names.add(target_layer_name)

            if not should_consider_scope(node.node_name, ignored_scopes=self.ignored_scopes):
                continue

            if node.metatype in OUTPUT_NOOP_METATYPES:
                continue

            is_custom, layer_info = converter.get_layer_info_for_node(node.node_name)
            if node.metatype in SPARSITY_LAYER_METATYPES:
                # Processing a regular weighted node
                for weight_def in node.metatype.weight_definitions:
                    op_name = self._get_sparsity_operation_name(node.node_name, weight_def.weight_attr_name)
                    self._op_names.append(op_name)

                    transformations.register(
                        TFInsertionCommand(
                            target_point=TFLayerWeight(layer_info.layer_name, weight_def.weight_attr_name),
                            callable_object=BinaryMask(op_name),
                            priority=TransformationPriority.SPARSIFICATION_PRIORITY,
                        )
                    )
            elif node.metatype in WEIGHTABLE_TF_OP_METATYPES:
                assert is_custom
                # Processing a custom layer weighted node
                # Caution: here layer_name will refer to the weight itself, not to the op
                weight_attr_name = node.layer_name
                op_name = self._get_sparsity_operation_name(node.node_name, weight_attr_name)
                self._op_names.append(op_name)

                transformations.register(
                    TFInsertionCommand(
                        target_point=TFLayerWeight(layer_info.layer_name, weight_attr_name),
                        callable_object=BinaryMaskWithWeightsBackup(op_name, weight_attr_name),
                        priority=TransformationPriority.SPARSIFICATION_PRIORITY,
                    )
                )

        return transformations

    def _get_sparsity_operation_name(self, layer_name: str, weight_attr_name: str) -> str:
        return f"{layer_name}_{weight_attr_name}_sparsity_binary_mask"

    def _build_controller(self, model: tf.keras.Model) -> "MagnitudeSparsityController":
        """
        Simple implementation of building controller without setting builder state and loading controller's one.
        Should be called once the compressed model target_model is fully constructed.

        :param model: The model with additional modifications necessary to enable
            algorithm-specific compression during fine-tuning.
        :return: The instance of the `MagnitudeSparsityController`.
        """
        return MagnitudeSparsityController(model, self.config, self._op_names)

    def initialize(self, model: tf.keras.Model) -> None:
        pass


@api()
@ADAPTIVE_COMPRESSION_CONTROLLERS.register("tf_magnitude_sparsity")
class MagnitudeSparsityController(BaseSparsityController):
    """
    Controller class for magnitude sparsity in TF.
    """

    def __init__(self, target_model, config: NNCFConfig, op_names):
        super().__init__(target_model, op_names)
        algo_config = extract_algo_specific_config(config, "magnitude_sparsity")
        params = deepcopy(algo_config.get("params", {}))
        self._threshold = 0
        self._frozen = False
        self._weight_importance_fn = WEIGHT_IMPORTANCE_FUNCTIONS[
            params.get("weight_importance", MAGNITUDE_SPARSITY_WEIGHT_IMPORTANCE)
        ]

        sparsity_init = algo_config.get("sparsity_init", SPARSITY_INIT)
        params["sparsity_init"] = sparsity_init
        scheduler_type = params.get("schedule", "polynomial")

        if scheduler_type == "adaptive":
            raise ValueError("Magnitude sparsity algorithm do not support adaptive scheduler")

        scheduler_cls = SPARSITY_SCHEDULERS.get(scheduler_type)
        self._scheduler: SparsityScheduler = scheduler_cls(self, params)
        self._loss = TFZeroCompressionLoss()
        self._bn_adaptation = None
        self._config = config
        self.set_sparsity_level(sparsity_init)

    @property
    def scheduler(self) -> SparsityScheduler:
        return self._scheduler

    @property
    def current_sparsity_level(self) -> float:
        return self._scheduler.current_sparsity_level

    @property
    def loss(self) -> CompressionLoss:
        return self._loss

    def freeze(self, freeze: bool = True):
        self._frozen = freeze

    def set_sparsity_level(self, sparsity_level, run_batchnorm_adaptation: bool = False):
        """
        Sets the sparsity level that should be applied to the model's weights.

        :param sparsity_level: Sparsity level that should be applied to the model's weights.
        :param run_batchnorm_adaptation: Whether to run batchnorm adaptation after setting the sparsity level.
        """
        if not self._frozen:
            if sparsity_level >= 1 or sparsity_level < 0:
                raise AttributeError(
                    "Sparsity level should be within interval [0,1), actual value to set is: {}".format(sparsity_level)
                )

            self._threshold = self._select_threshold(sparsity_level)
            self._set_masks_for_threshold(self._threshold)

        if run_batchnorm_adaptation:
            self._run_batchnorm_adaptation()

    def _select_threshold(self, sparsity_level):
        all_weights = self._collect_all_weights()
        if not all_weights:
            return 0.0
        all_weights_tensor = tf.sort(tf.concat(all_weights, 0))
        index = int(tf.cast(tf.size(all_weights_tensor) - 1, all_weights_tensor.dtype) * sparsity_level)
        threshold = all_weights_tensor[index].numpy()
        return threshold

    def _set_masks_for_threshold(self, threshold_val):
        for wrapped_layer in collect_wrapped_layers(self._model):
            for weight_attr, ops in wrapped_layer.weights_attr_ops.items():
                weight = wrapped_layer.layer_weights[weight_attr]

                for op_name in ops:
                    if op_name in self._op_names:
                        wrapped_layer.ops_weights[op_name]["mask"].assign(
                            calc_magnitude_binary_mask(weight, self._weight_importance_fn, threshold_val)
                        )

    def _collect_all_weights(self):
        all_weights = []
        for wrapped_layer in collect_wrapped_layers(self._model):
            for weight_attr, ops in wrapped_layer.weights_attr_ops.items():
                for op_name in ops:
                    if op_name in self._op_names:
                        all_weights.append(
                            tf.reshape(self._weight_importance_fn(wrapped_layer.layer_weights[weight_attr]), [-1])
                        )
        return all_weights

    @property
    def compression_rate(self) -> float:
        return self.statistics().magnitude_sparsity.model_statistics.sparsity_level

    @compression_rate.setter
    def compression_rate(self, compression_rate: float) -> None:
        self.freeze(False)
        self.set_sparsity_level(compression_rate)
        self.freeze(True)

    def disable_scheduler(self):
        self._scheduler = StubCompressionScheduler()
        self._scheduler.target_level = 0.0
        self._scheduler.current_sparsity_level = 0.0

    def statistics(self, quickly_collected_only: bool = False) -> NNCFStatistics:
        collector = TFSparseModelStatisticsCollector(self.model, self._op_names)
        model_stats = collector.collect()

        threshold_stats = []
        threshold = self._select_threshold(model_stats.sparsity_level)
        for s in model_stats.sparsified_layers_summary:
            threshold_stats.append(LayerThreshold(s.name, threshold))

        target_sparsity_level = self.scheduler.current_sparsity_level

        stats = MagnitudeSparsityStatistics(model_stats, threshold_stats, target_sparsity_level)

        nncf_stats = NNCFStatistics()
        nncf_stats.register("magnitude_sparsity", stats)
        return nncf_stats

    def compression_stage(self) -> CompressionStage:
        if self.scheduler.current_sparsity_level >= self.scheduler.target_level:
            return CompressionStage.FULLY_COMPRESSED
        if self.scheduler.current_sparsity_level == 0:
            return CompressionStage.UNCOMPRESSED
        return CompressionStage.PARTIALLY_COMPRESSED

    def _run_batchnorm_adaptation(self):
        if self._bn_adaptation is None:
            self._bn_adaptation = BatchnormAdaptationAlgorithm(
                **extract_bn_adaptation_init_params(self._config, self.name)
            )
        self._bn_adaptation.run(self.model)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial
from typing import Any, Callable, Dict, List, Mapping, Sequence, Set, Tuple, Type, Union

string_types = (str, bytes)
iteritems = lambda mapping: getattr(mapping, "iteritems", mapping.items)()


def to_tuple(lst: List, named_tuple_class: Type = None, named_tuple_fields: List[str] = None) -> Tuple:
    # Able to produce namedtuples if a corresponding parameter is given
    if named_tuple_fields is None:
        return tuple(lst)
    return named_tuple_class(*lst)


def is_tuple(obj) -> bool:
    return isinstance(obj, tuple)


def is_named_tuple(obj) -> bool:
    return is_tuple(obj) and (obj.__class__ is not tuple)


def maybe_get_iterator(obj):
    it = None

    if isinstance(obj, Mapping):
        it = iteritems

    elif isinstance(obj, (Sequence, Set)) and not isinstance(obj, string_types):
        it = enumerate
    return it


class InputIndexEntry:
    def __init__(self, path: Tuple[Union[int, str], ...], getter: Callable, setter: Callable):
        self.path = path
        self.getter = getter
        self.setter = setter


class TupleRebuildingSetter:
    def __init__(self, idx_to_set, current_tuple, previous_level_setter_for_current_tuple):
        self._previous_level_setter = previous_level_setter_for_current_tuple
        self._current_tuple = current_tuple
        self._idx_to_set = idx_to_set

    def __call__(self, value):
        tmp_list = list(self._current_tuple)
        tmp_list[self._idx_to_set] = value
        new_tuple = tuple(tmp_list)
        self._current_tuple = new_tuple
        self._previous_level_setter(new_tuple)


class NestedObjectIndex:
    def __init__(self, obj, path=(), memo=None, previous_level_setter=None):
        self._flat_nested_obj_indexing: List[InputIndexEntry] = []
        self._nested_object_paths_generator(obj, self._flat_nested_obj_indexing, path, memo, previous_level_setter)

    @staticmethod
    def _nested_object_paths_generator(obj, out_entries_list, path=(), memo=None, previous_level_setter=None):
        if memo is None:
            memo = set()
        iterator = maybe_get_iterator(obj)
        if iterator is not None:
            if id(obj) not in memo:
                memo.add(id(obj))
                current_level_getters = []
                current_level_setters = []
                for idx, iterval in enumerate(iterator(obj)):
                    path_component, value = iterval
                    current_level_getters.append(partial(obj.__getitem__, path_component))
                    if not isinstance(obj, tuple):
                        # `range` objects, for instance, have no __setitem__ and should be disregarded
                        if hasattr(obj, "__setitem__"):
                            current_level_setters.append(partial(obj.__setitem__, path_component))
                        else:
                            current_level_setters.append(None)
                    else:
                        current_level_setters.append(TupleRebuildingSetter(idx, obj, previous_level_setter))

                for idx, iterval in enumerate(iterator(obj)):
                    path_component, value = iterval
                    retval = NestedObjectIndex._nested_object_paths_generator(
                        value, out_entries_list, path + (path_component,), memo, current_level_setters[idx]
                    )
                    was_leaf = retval[1]
                    if was_leaf:
                        leaf_entry_path = retval
                        # getter = partial(obj.__getitem__, path_component)
                        getter = current_level_getters[idx]
                        setter = current_level_setters[idx]
                        if setter is not None:  # see note above about non-settable objects
                            out_entries_list.append(InputIndexEntry(leaf_entry_path, getter, setter))

                memo.remove(id(obj))
            is_leaf = False
            return path, is_leaf

        is_leaf = True
        return path, is_leaf

    def get_flat_nested_obj_indexing(self) -> List[InputIndexEntry]:
        return self._flat_nested_obj_indexing


def objwalk(obj, unary_predicate: Callable[[Any], bool], apply_fn: Callable, memo=None):
    """
    Walks through the indexable container hierarchy of obj and replaces all sub-objects matching a criterion
    with the result of a given function application.
    """

    if memo is None:
        memo = set()

    named_tuple_class = None
    named_tuple_fields = None
    if is_named_tuple(obj):
        named_tuple_class = obj.__class__

        named_tuple_fields = obj._fields

    was_tuple = is_tuple(obj)
    if was_tuple:
        obj = list(obj)

    iterator = maybe_get_iterator(obj)

    if iterator is not None:
        if id(obj) not in memo:
            memo.add(id(obj))
            indices_to_apply_fn_to = set()
            indices_vs_named_tuple_data: Dict[Any, Tuple[list, Type, List[str]]] = {}
            for idx, value in iterator(obj):
                next_level_it = maybe_get_iterator(value)
                if next_level_it is None:
                    if unary_predicate(value):
                        indices_to_apply_fn_to.add(idx)
                else:
                    if is_tuple(value):
                        processed_tuple = objwalk(value, unary_predicate, apply_fn, memo)
                        if is_named_tuple(value):
                            indices_vs_named_tuple_data[idx] = processed_tuple, value.__class__, value._fields
                        else:
                            indices_vs_named_tuple_data[idx] = processed_tuple, None, None
                    else:
                        objwalk(value, unary_predicate, apply_fn)
            for idx in indices_to_apply_fn_to:
                obj[idx] = apply_fn(obj[idx])
            for idx, tpl_data in indices_vs_named_tuple_data.items():
                tpl, n_tpl_class, n_tpl_fields = tpl_data
                obj[idx] = to_tuple(tpl, n_tpl_class, n_tpl_fields)

            memo.remove(id(obj))
    else:
        if unary_predicate(obj):
            return apply_fn(obj)

    if was_tuple:
        return to_tuple(obj, named_tuple_class, named_tuple_fields)

    return obj



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Callable, Iterable, List, Optional, Tuple, TypeVar, Union

import nncf
from nncf.api.compression import TModel
from nncf.common.deprecation import warning_deprecated
from nncf.common.graph import NNCFGraph
from nncf.common.graph.operator_metatypes import OperatorMetatype
from nncf.common.logging.logger import nncf_logger
from nncf.common.quantization.structs import QuantizationPreset
from nncf.common.utils.api_marker import api
from nncf.common.utils.backend import BackendType
from nncf.common.utils.backend import get_backend
from nncf.data import Dataset
from nncf.parameters import CompressWeightsMode
from nncf.parameters import DropType
from nncf.parameters import ModelType
from nncf.parameters import QuantizationMode
from nncf.parameters import SensitivityMetric
from nncf.parameters import TargetDevice
from nncf.quantization.advanced_parameters import AdvancedAccuracyRestorerParameters
from nncf.quantization.advanced_parameters import AdvancedCompressionParameters
from nncf.quantization.advanced_parameters import AdvancedQuantizationParameters
from nncf.quantization.algorithms.accuracy_control.evaluator import MetricResults
from nncf.quantization.algorithms.hyperparameter_tuner.algorithm import HyperparameterTuner
from nncf.quantization.algorithms.hyperparameter_tuner.param_grid import get_quantization_param_grids
from nncf.quantization.algorithms.post_training.pipeline import create_ptq_pipeline
from nncf.scopes import IgnoredScope

TTensor = TypeVar("TTensor")

BATCHWISE_STATISTICS_WARNING = (
    "For the particular model the batchwise statistics collection can lead to inaccurate statistics. "
    "If the accuracy degradation after compression is unsatisfactory, then "
    "the recomendation is to turn off batchwise statistics. If the results are still unsatisfactory, "
    "provide a dataloader with batch_size = 1 to the calibration dataset."
)


def warning_model_no_batchwise_support(
    graph: NNCFGraph,
    advanced_quantization_parameters: Optional[AdvancedQuantizationParameters],
    model_type: ModelType,
    no_batchwise_support_metatypes: List[OperatorMetatype],
) -> None:
    """
    Prints the warning message if batchwise statistics could lead to a significant accuracy drop.

    :param graph: Model's NNCFGraph.
    :param advanced_quantization_parameters: AdvancedQuantizationParameters.
    :param model_type: Model type algorithm option.
    :param no_batchwise_support_metatypes: Meatypes having no batchwise statistics support.
    """
    if (
        advanced_quantization_parameters
        and advanced_quantization_parameters.batchwise_statistics
        and (graph.get_nodes_by_metatypes(no_batchwise_support_metatypes) or model_type == ModelType.TRANSFORMER)
    ):
        nncf_logger.warning(BATCHWISE_STATISTICS_WARNING)


def _update_advanced_quantization_parameters(
    advanced_parameters: Optional[AdvancedQuantizationParameters], calibration_dataset: Dataset
) -> AdvancedQuantizationParameters:
    """
    Updates AdvancedQuantizationParameters depending on batch_size.

    :param advanced_parameters: Advanced quantization parameters for
        fine-tuning the quantization algorithm.
    :param calibration_dataset: A representative dataset for the
        calibration process.
    :return: Updated AdvancedQuantizationParameters.
    """
    batch_size = calibration_dataset.get_batch_size()
    if batch_size is not None and batch_size > 1:
        if advanced_parameters is None:
            advanced_parameters = AdvancedQuantizationParameters(batchwise_statistics=True)
        elif advanced_parameters.batchwise_statistics is None:
            advanced_parameters.batchwise_statistics = True
    return advanced_parameters


@api(canonical_alias="nncf.quantize")
def quantize(
    model: TModel,
    calibration_dataset: Dataset,
    mode: Optional[QuantizationMode] = None,
    preset: Optional[QuantizationPreset] = None,
    target_device: TargetDevice = TargetDevice.ANY,
    subset_size: int = 300,
    fast_bias_correction: bool = True,
    model_type: Optional[ModelType] = None,
    ignored_scope: Optional[IgnoredScope] = None,
    advanced_parameters: Optional[AdvancedQuantizationParameters] = None,
) -> TModel:
    """
    Applies post-training quantization to the provided model.

    :param model: A model to be quantized.
    :type  model: TModel
    :param calibration_dataset: A representative dataset for the
        calibration process.
    :type  calibration_dataset: nncf.Dataset
    :param mode: Special quantization mode that specify different ways of the optimization.
    :type mode: Optional[nncf.QuantizationMode]
    :param preset: A preset controls the quantization mode (symmetric and asymmetric).
        It can take the following values:
        - `performance`: Symmetric quantization of weights and activations.
        - `mixed`: Symmetric quantization of weights and asymmetric quantization of activations.
        Default value is None. In this case, `mixed` preset is used for `transformer`
        model type otherwise `performance`.
    :type  preset: nncf.QuantizationPreset
    :param target_device: A target device the specificity of which will be taken
        into account while compressing in order to obtain the best performance
        for this type of device.
    :type  target_device: nncf.TargetDevice
    :param subset_size: Size of a subset to calculate activations statistics used for quantization.
        Must be positive.
    :param fast_bias_correction: Setting this option to `False` enables a different
        bias correction method which is more accurate, in general, and takes
        more time but requires less memory.
    :param model_type: Model type is needed to specify additional patterns
        in the model. Supported only `transformer` now.
    :type  model_type: Optional[nncf.ModelType]
    :param ignored_scope: An ignored scope that defined the list of model control
        flow graph nodes to be ignored during quantization.
    :type  ignored_scope: Optional[nncf.IgnoredScope]
    :param advanced_parameters: Advanced quantization parameters for
        fine-tuning the quantization algorithm.
    :return: The quantized model.
    :rtype: TModel
    """
    if subset_size < 1:
        raise ValueError("Subset size must be positive.")

    advanced_parameters = _update_advanced_quantization_parameters(advanced_parameters, calibration_dataset)

    backend = get_backend(model)
    if backend == BackendType.OPENVINO:
        from nncf.openvino.quantization.quantize_model import quantize_impl

        return quantize_impl(
            model=model,
            calibration_dataset=calibration_dataset,
            mode=mode,
            preset=preset,
            target_device=target_device,
            subset_size=subset_size,
            fast_bias_correction=fast_bias_correction,
            model_type=model_type,
            ignored_scope=ignored_scope,
            advanced_parameters=advanced_parameters,
        )

    if backend == BackendType.ONNX:
        from nncf.onnx.quantization.quantize_model import quantize_impl

        return quantize_impl(
            model=model,
            calibration_dataset=calibration_dataset,
            mode=mode,
            preset=preset,
            target_device=target_device,
            subset_size=subset_size,
            fast_bias_correction=fast_bias_correction,
            model_type=model_type,
            ignored_scope=ignored_scope,
            advanced_parameters=advanced_parameters,
        )

    if backend == BackendType.TENSORFLOW:
        from nncf.tensorflow.quantization.quantize_model import quantize_impl

        return quantize_impl(
            model=model,
            calibration_dataset=calibration_dataset,
            mode=mode,
            preset=preset,
            target_device=target_device,
            subset_size=subset_size,
            fast_bias_correction=fast_bias_correction,
            model_type=model_type,
            ignored_scope=ignored_scope,
            advanced_parameters=advanced_parameters,
        )

    if backend == BackendType.TORCH:
        from nncf.torch.quantization.quantize_model import quantize_impl

        return quantize_impl(
            model=model,
            calibration_dataset=calibration_dataset,
            mode=mode,
            preset=preset,
            target_device=target_device,
            subset_size=subset_size,
            fast_bias_correction=fast_bias_correction,
            model_type=model_type,
            ignored_scope=ignored_scope,
            advanced_parameters=advanced_parameters,
        )

    raise nncf.UnsupportedBackendError(f"Unsupported type of backend: {backend}")


@api(canonical_alias="nncf.quantize_with_accuracy_control")
def quantize_with_accuracy_control(
    model: TModel,
    calibration_dataset: Dataset,
    validation_dataset: Dataset,
    validation_fn: Callable[[Any, Iterable[Any]], float],
    max_drop: float = 0.01,
    drop_type: DropType = DropType.ABSOLUTE,
    preset: Optional[QuantizationPreset] = None,
    target_device: TargetDevice = TargetDevice.ANY,
    subset_size: int = 300,
    fast_bias_correction: bool = True,
    model_type: Optional[ModelType] = None,
    ignored_scope: Optional[IgnoredScope] = None,
    advanced_quantization_parameters: Optional[AdvancedQuantizationParameters] = None,
    advanced_accuracy_restorer_parameters: Optional[AdvancedAccuracyRestorerParameters] = None,
) -> TModel:
    """
    Applies post-training quantization algorithm with accuracy control to provided model.

    :param model: A model to be quantized.
    :type model: TModel
    :param calibration_dataset: A representative dataset for the calibration process.
    :type calibration_dataset: nncf.Dataset
    :param validation_dataset: A dataset for the validation process.
    :type validation_dataset: nncf.Dataset
    :param validation_fn: A validation function to validate the model. It should take two arguments:
        - `model`: model to be validate.
        - `validation_dataset`: dataset that provides data items to
              validate the provided model.
        The function should return the value of the metric with the following meaning:
        A higher value corresponds to better performance of the model.
    :param max_drop: The maximum accuracy drop that should be achieved after the quantization.
    :param drop_type: The accuracy drop type, which determines how the maximum accuracy
        drop between the original model and the compressed model is calculated.
    :param preset: A preset controls the quantization mode (symmetric and asymmetric).
        It can take the following values:
        - `performance`: Symmetric quantization of weights and activations.
        - `mixed`: Symmetric quantization of weights and asymmetric quantization of activations.
        Default value is None. In this case, `mixed` preset is used for `transformer`
        model type otherwise `performance`.
    :type preset: nncf.QuantizationPreset
    :param target_device: A target device the specificity of which will be taken
        into account while compressing in order to obtain the best performance
        for this type of device.
    :type target_device: nncf.TargetDevice
    :param subset_size: Size of a subset to calculate activations
        statistics used for quantization.
    :param fast_bias_correction: Setting this option to `False` enables a different
        bias correction method which is more accurate, in general, and takes
        more time but requires less memory.
    :param model_type: Model type is needed to specify additional patterns
        in the model. Supported only `transformer` now.
    :type model_type: nncf.ModelType
    :param ignored_scope: An ignored scope that defined the list of model control
        flow graph nodes to be ignored during quantization.
    :type ignored_scope: nncf.IgnoredScope
    :param advanced_quantization_parameters: Advanced quantization parameters for
        fine-tuning the quantization algorithm.
    :param advanced_accuracy_restorer_parameters: Advanced parameters for fine-tuning
        the accuracy restorer algorithm.
    :type advanced_accuracy_restorer_parameters: Optional[AdvancedAccuracyRestorerParameters]
    :return: The quantized model.
    :rtype: TModel
    """
    advanced_quantization_parameters = _update_advanced_quantization_parameters(
        advanced_quantization_parameters, calibration_dataset
    )

    backend = get_backend(model)
    if backend == BackendType.OPENVINO:
        from nncf.openvino.quantization.quantize_model import quantize_with_accuracy_control_impl

        return quantize_with_accuracy_control_impl(
            model,
            calibration_dataset,
            validation_dataset,
            validation_fn,
            max_drop,
            drop_type,
            preset,
            target_device,
            subset_size,
            fast_bias_correction,
            model_type,
            ignored_scope,
            advanced_quantization_parameters,
            advanced_accuracy_restorer_parameters,
        )
    if backend == BackendType.ONNX:
        from nncf.onnx.quantization.quantize_model import quantize_with_accuracy_control_impl

        return quantize_with_accuracy_control_impl(
            model,
            calibration_dataset,
            validation_dataset,
            validation_fn,
            max_drop,
            drop_type,
            preset,
            target_device,
            subset_size,
            fast_bias_correction,
            model_type,
            ignored_scope,
            advanced_quantization_parameters,
            advanced_accuracy_restorer_parameters,
        )

    raise nncf.UnsupportedBackendError(f"Unsupported type of backend: {backend}")


@api(canonical_alias="nncf.compress_weights")
def compress_weights(
    model: TModel,
    mode=CompressWeightsMode.INT8_ASYM,
    ratio: Optional[float] = None,
    group_size: Optional[int] = None,
    ignored_scope: Optional[IgnoredScope] = None,
    all_layers: Optional[bool] = None,
    dataset: Optional[Dataset] = None,
    sensitivity_metric: Optional[SensitivityMetric] = None,
    *,
    subset_size: Optional[int] = 128,
    awq: Optional[bool] = None,
    scale_estimation: Optional[bool] = None,
    gptq: Optional[bool] = None,
    advanced_parameters: Optional[AdvancedCompressionParameters] = None,
) -> TModel:
    """
    Compress model weights.

    :param model: A model to be compressed.
    :type model: TModel
    :param mode: Defines a mode for weight compression.
        INT8_SYM stands for 8-bit integer symmetric quantization of all weights.
        INT8_ASYM is the same as INT8_SYM mode, but weights are quantized to a primary precision asymmetrically
            with a typical non-fixed zero point.
        INT4_SYM stands for a mixed-precision weights quantization with 4-bit integer as a primary precision.
            Weights are quantized to a primary precision symmetrically with a fixed zero point equals to 8.
            All embeddings and the last layer are always compressed to a backup precision, which is INT8_ASYM,
            by default. All others are quantized whether to 4-bit integer or to a backup precision depending on
            criteria and the given ratio.
        INT4_ASYM is the same as INT4_SYM mode, but weights are quantized to a primary precision asymmetrically
            with a typical non-fixed zero point.
        NF4 is the same as INT4_SYM mode, but primary precision is NF4 data type without zero point.
    :type mode: nncf.CompressWeightsMode
    :param ratio: the ratio between baseline and backup precisions (e.g. 0.9 means 90% of layers quantized to NF4
        and the rest to INT8_ASYM).
    :type ratio: float
    :param group_size: number of weights (e.g. 128) in the channel dimension that share quantization parameters (scale).
        The value -1 means no grouping.
    :type group_size: int
    :param ignored_scope: An ignored scope that defined the list of model control
        flow graph nodes to be ignored during quantization.
    :type ignored_scope: nncf.IgnoredScope
    :param all_layers: Indicates whether embeddings and last MatMul layers should be compressed to a primary
        precision. By default, the backup precision is assigned for the embeddings and last MatMul layers.
    :type all_layers: bool
    :param dataset: Dataset used for assigning different quantization precision by finding outliers in activations.
    :type dataset: nncf.Dataset
    :param sensitivity_metric: The sensitivity metric for assigning quantization precision to layers. In order to
        preserve the accuracy of the model, the more sensitive layers receives a higher precision.
    :type sensitivity_metric: nncf.SensitivityMetric
    :param subset_size: Number of data samples to calculate activation statistics used for assigning different
        quantization precision. Defaults to 128.
    :type subset_size: int
    :param awq: Indicates whether use AWQ weights correction.
    :type awq: bool
    :param scale_estimation: Indicates whether a scale estimation algorithm is used that minimizes the L2 error
        between the original and compressed layers.
    :type scale_estimation: bool
    :param gptq: Indicates whether use GPTQ algorithm.
    :type gptq: bool
    :param advanced_parameters: Advanced parameters for compression algorithms.
    :type advanced_parameters: nncf.AdvancedCompressionParameters
    :return: The non-trainable model with compressed weights.
    """
    if mode == CompressWeightsMode.INT8:
        warning_deprecated(
            "`CompressWeightsMode.INT8` is deprecated." "Please, use `CompressWeightsMode.INT8_ASYM` as value instead."
        )
        mode = CompressWeightsMode.INT8_ASYM

    backend = get_backend(model)
    compression_weights_impl = None

    if backend == BackendType.TORCH:
        from nncf.torch.model_creation import is_wrapped_model
        from nncf.torch.model_creation import wrap_model
        from nncf.torch.quantization.quantize_model import compress_weights_impl as pt_compression_weights_impl

        if mode not in [CompressWeightsMode.INT8_ASYM, CompressWeightsMode.INT8_SYM]:
            raise AttributeError(
                "Torch backend supports only INT8_ASYM, INT8_SYM modes for weight compression, "
                f"but given {mode.value} mode."
            )

        if True in [awq, scale_estimation, gptq]:
            raise AttributeError(
                "Torch backend doesn`t supports scale estimation and AWQ algorithm, "
                "but awq=True or scale_estimation=True or gptq=True is specified."
            )

        if is_wrapped_model(model):
            if not model.nncf.trace_parameters:
                raise ValueError(
                    "Tracing capabilities with tracing parameters are required in the PyTorch model "
                    "for nncf.compress_weights(). Please wrap the model using "
                    "nncf.torch.wrap_model(model, example_input, trace_parameters=True) before calling "
                    "nncf.compress_weights()."
                )
        elif dataset is None:
            raise AttributeError("Please provide a dataset of at least one element for PyTorch model tracing.")
        else:
            example_input = next(iter(dataset.get_inference_data()))
            model = wrap_model(model, example_input=example_input, trace_parameters=True)
        dataset = None
        compression_weights_impl = pt_compression_weights_impl

    if backend == BackendType.OPENVINO:
        from nncf.openvino.quantization.quantize_model import compress_weights_impl as ov_compress_weights_impl

        if any((awq, scale_estimation)) and (dataset is None or mode == CompressWeightsMode.NF4 or group_size == -1):
            raise AttributeError(
                "Scale estimation or AWQ algorithm defined, but dataset is None or mode is NF4 or group_size < 0."
            )
        if gptq and (dataset is None or group_size == -1):
            raise AttributeError("GPTQ algorithm defined, but dataset is None or group_size < 0.")

        if gptq and scale_estimation:
            raise AttributeError(
                "Simultaneous use of Scale estimation and GPTQ algorithms is not supported. Select one of them."
            )

        compression_weights_impl = ov_compress_weights_impl

    if mode in [CompressWeightsMode.INT8_ASYM, CompressWeightsMode.INT8_SYM]:
        if ratio is None:
            ratio = 1
        if group_size is None:
            group_size = -1
        if ratio != 1 or group_size != -1:
            raise AttributeError(
                "INT8 mode assumes per-channel quantization of all layers in 8 bit. "
                "Default values of `ratio` (1) and `group_size` (-1) parameters can not be overridden"
            )
        options = [all_layers, sensitivity_metric, dataset, awq, scale_estimation, gptq]
        if any(option is not None for option in options):
            raise AttributeError(
                "INT8 modes do not support `all_layers`, `sensitivity_metric`, `awq`, `scale_estimation`, `gptq` "
                "and `dataset` options. Set them to None."
            )

    if ratio is None:
        ratio = 1
    if group_size is None:
        group_size = 128
    if all_layers is None:
        all_layers = False
    if awq is None:
        awq = False
    if scale_estimation is None:
        scale_estimation = False
    if gptq is None:
        gptq = False
    if ignored_scope is None:
        ignored_scope = IgnoredScope()
    if sensitivity_metric is None:
        sensitivity_metric = (
            SensitivityMetric.WEIGHT_QUANTIZATION_ERROR
            if dataset is None
            else SensitivityMetric.MAX_ACTIVATION_VARIANCE
        )
    if ratio != 1 and dataset is None and sensitivity_metric != SensitivityMetric.WEIGHT_QUANTIZATION_ERROR:
        raise AttributeError(
            f"Mixed precision selection based on the given sensitivity metric={sensitivity_metric.value} requires "
            "a dataset, but it's not provided."
        )
    if ratio < 0 or ratio > 1:
        raise ValueError(f"The ratio should be between 0 and 1, but ratio={ratio} is specified.")
    if subset_size is None or subset_size <= 0:
        raise ValueError(f"The subset_size value should be positive, but subset_size={subset_size} is given.")

    if compression_weights_impl is None:
        raise nncf.UnsupportedBackendError(f"Unsupported type of backend: {backend}")

    return compression_weights_impl(
        model,
        dataset,
        mode,
        ratio,
        group_size,
        ignored_scope,
        all_layers,
        sensitivity_metric,
        awq,
        subset_size,
        scale_estimation,
        gptq,
        advanced_parameters,
    )


def quantize_with_tune_hyperparams(
    model: TModel,
    calibration_dataset: Dataset,
    validation_dataset: Dataset,
    validation_fn: Callable[[Any, Iterable[Any]], Tuple[float, Union[None, List[float], List[List[TTensor]]]]],
    initial_metric_results: MetricResults,
    quantized_metric_results: MetricResults,
    tuner_subset_size: int = 300,
    preset: Optional[QuantizationPreset] = None,
    target_device: TargetDevice = TargetDevice.ANY,
    subset_size: int = 300,
    fast_bias_correction: bool = True,
    model_type: Optional[ModelType] = None,
    ignored_scope: Optional[IgnoredScope] = None,
    advanced_quantization_parameters: Optional[AdvancedQuantizationParameters] = None,
) -> TModel:
    """
    Applies post-training quantization algorithm with tune hyperparameters to provided model.

    :param model: A model to be quantized.
    :param calibration_dataset: A representative dataset for the calibration process.
    :param validation_dataset: : A dataset for the validation process.
    :param validation_fn: A validation function to validate the model.
    :param initial_metric_results: Initial metric results.
    :param quantized_metric_results: Quantized metric results.
    :param tuner_subset_size: Tuner subset size.
    :param preset: A preset controls the quantization mode (symmetric and asymmetric).
        It can take the following values:
        - `performance`: Symmetric quantization of weights and activations.
        - `mixed`: Symmetric quantization of weights and asymmetric quantization of activations.
        Default value is None. In this case, `mixed` preset is used for `transformer`
        model type otherwise `performance`.
    :param target_device: A target device the specificity of which will be taken
        into account while compressing in order to obtain the best performance
        for this type of device.
    :param subset_size: Size of a subset to calculate activations
        statistics used for quantization.
    :param fast_bias_correction: Setting this option to `False` enables a different
        bias correction method which is more accurate, in general, and takes
        more time but requires less memory.
    :param model_type: Model type is needed to specify additional patterns
        in the model. Supported only `transformer` now.
    :param ignored_scope: An ignored scope that defined the list of model control
        flow graph nodes to be ignored during quantization.
    :param advanced_quantization_parameters: Advanced quantization parameters for
        fine-tuning the quantization algorithm.
    :return: The quantized model.
    """
    init_quantization_params = {
        "preset": preset,
        "target_device": target_device,
        "subset_size": subset_size,
        "fast_bias_correction": fast_bias_correction,
        "model_type": model_type,
        "ignored_scope": ignored_scope,
        "advanced_parameters": advanced_quantization_parameters,
    }

    backend = get_backend(model)
    param_grids = get_quantization_param_grids(create_ptq_pipeline(**init_quantization_params), backend)

    hyperparameter_tuner = HyperparameterTuner(
        create_ptq_pipeline,
        init_quantization_params,
        param_grids,
        calibration_dataset,
        validation_fn,
        tuner_subset_size,
        initial_metric_results,
        quantized_metric_results,
    )

    quantized_model = hyperparameter_tuner.apply(model, validation_dataset)

    return quantized_model



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from nncf.torch.external_hook import ExternalOpCallHook
from nncf.torch.quantization.debug_interface import QuantizationDebugInterface

EXTERNAL_QUANTIZERS_STORAGE_NAME = "external_quantizers"
EXTERNAL_QUANTIZERS_STORAGE_PREFIX = "_nncf." + EXTERNAL_QUANTIZERS_STORAGE_NAME


class ExternalQuantizerCallHook(ExternalOpCallHook):
    """
    External hook which is using quantization storage name and
    could be constructed with a debug interface.
    """

    def __init__(
        self,
        quantizer_storage_key: str,
        debug_interface: QuantizationDebugInterface = None,
    ):
        super().__init__(EXTERNAL_QUANTIZERS_STORAGE_NAME, quantizer_storage_key)
        self.debug_interface = debug_interface

    def __call__(self, *args, **kwargs):
        if self.debug_interface is not None:
            self.debug_interface.register_activation_quantize_call(str(self._storage_key))
        return super().__call__(*args, **kwargs)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from dataclasses import field
from enum import Enum
from typing import Optional

from nncf.common.utils.api_marker import api


@api()
class StatisticsType(Enum):
    """
    Enumeration of different types of statistics that are used to collect per sample
    statistics for activations and weights of the model.

    :param MAX: The maximum value in a tensor.
    :param MIN: The minimum value in a tensor.
    :param ABS_MAX: The maximum absolute value in a tensor.
    :param QUANTILE: A specific quantile value in a tensor.
    :param ABS_QUANTILE: A specific quantile value in the absolute tensor.
    :param MEAN: The mean value of a tensor.
    """

    MAX = "max"
    MIN = "min"
    ABS_MAX = "abs_max"
    QUANTILE = "quantile"
    ABS_QUANTILE = "abs_quantile"
    MEAN = "mean"


@api()
class AggregatorType(Enum):
    """
    Enumeration of different types of aggregators that are used to aggregate per sample
    statistics for activations and weights of the model.

    :param MEAN: The mean value of a set of tensors.
    :param MAX: The maximum value of a set of tensors.
    :param MIN: The minimum value of a set of tensors.
    :param MEDIAN: The median value of a set of tensors.
    :param MEAN_NO_OUTLIERS: The mean value of a set of tensors with outliers removed.
    :param MEDIAN_NO_OUTLIERS: The median value of a set of tensors with outliers removed.
    """

    MEAN = "mean"
    MAX = "max"
    MIN = "min"
    MEDIAN = "median"
    MEAN_NO_OUTLIERS = "mean_no_outliers"
    MEDIAN_NO_OUTLIERS = "median_no_outliers"


@api()
@dataclass
class StatisticsCollectorParameters:
    """
    Contains parameters for collecting statistics for activations and weights of the model.

    :param statistics_type: The type of per sample statistics to collect.
    :type statistics_type: Optional[nncf.quantization.range_estimator.StatisticsType]
    :param aggregator_type: The type of aggregator of per sample statistics.
    :type aggregator_type: Optional[nncf.quantization.range_estimator.AggregatorType]
    :param clipping_value: The value to use for clipping the input tensors before
        collecting statistics.
    :type clipping_value: Optional[float]
    :param quantile_outlier_prob: The outlier probability for QUANTILE statistic or MEAN_NO_OUTLIERS/MEDIAN_NO_OUTLIERS
        aggregators. When using these together, please be aware that currently the quantile value will be applied to
        both.
    :type quantile_outlier_prob: float
    """

    statistics_type: Optional[StatisticsType] = None
    aggregator_type: Optional[AggregatorType] = None
    clipping_value: Optional[float] = None
    quantile_outlier_prob: float = 1e-4


@api()
@dataclass
class RangeEstimatorParameters:
    """
    Contains parameters for estimating the range of activations and weights of the model.

    :param min: The parameters for estimating the lower bound of the range.
    :type min: nncf.quantization.range_estimator.StatisticsCollectorParameters
    :param max: The Parameters for estimating the upper bound of the range.
    :type max: nncf.quantization.range_estimator.StatisticsCollectorParameters
    """

    min: StatisticsCollectorParameters = field(default_factory=StatisticsCollectorParameters)
    max: StatisticsCollectorParameters = field(default_factory=StatisticsCollectorParameters)


class RangeEstimatorParametersSet:
    """
    A class for specifying different sets of range estimator parameters.

    :param MINMAX: The range estimator parameters where the low bound of the range is
        calculated as global minimum of values of input tensors, the upper bound of
        the range as global maxima of the same values.
    :param MEAN_MINMAX: The range estimator parameters where the low bound of the range
        is calculated as average (across every sample) of minima of input tensors,
        the upper bound of the range as average of maxima of the same values.
    :param MEDIAN_MINMAX: The range estimator parameters where the low bound of the range
        is calculated as median (across every sample) of minima of input tensors,
        the upper bound of the range as median of maxima of the same values.
    :param MEAN_NO_OUTLIERS_MINMAX : The range estimator parameters where the low bound of the range
        is calculated as average (across all samples in range [min quantile, max quantile])
        of minima of input tensors, the upper bound of the range as average of maxima of the same values.
    :param MEAN_QUANTILE : The range estimator parameters where the low bound of the range
        is calculated as average (across every sample) of (quantile outlier probability)-quantiles,
        the upper bound of the range as average of (1 - quantile outlier probability)-quantiles of the same values.
    """

    MINMAX = RangeEstimatorParameters(
        min=StatisticsCollectorParameters(statistics_type=StatisticsType.MIN, aggregator_type=AggregatorType.MIN),
        max=StatisticsCollectorParameters(statistics_type=StatisticsType.MAX, aggregator_type=AggregatorType.MAX),
    )

    MEAN_MINMAX = RangeEstimatorParameters(
        min=StatisticsCollectorParameters(statistics_type=StatisticsType.MIN, aggregator_type=AggregatorType.MEAN),
        max=StatisticsCollectorParameters(statistics_type=StatisticsType.MAX, aggregator_type=AggregatorType.MEAN),
    )

    MEDIAN_MINMAX = RangeEstimatorParameters(
        min=StatisticsCollectorParameters(statistics_type=StatisticsType.MIN, aggregator_type=AggregatorType.MEDIAN),
        max=StatisticsCollectorParameters(statistics_type=StatisticsType.MAX, aggregator_type=AggregatorType.MEDIAN),
    )

    MEAN_NO_OUTLIERS_MINMAX = RangeEstimatorParameters(
        min=StatisticsCollectorParameters(
            statistics_type=StatisticsType.MIN, aggregator_type=AggregatorType.MEAN_NO_OUTLIERS
        ),
        max=StatisticsCollectorParameters(
            statistics_type=StatisticsType.MAX, aggregator_type=AggregatorType.MEAN_NO_OUTLIERS
        ),
    )

    MEAN_QUANTILE = RangeEstimatorParameters(
        min=StatisticsCollectorParameters(statistics_type=StatisticsType.QUANTILE, aggregator_type=AggregatorType.MEAN),
        max=StatisticsCollectorParameters(statistics_type=StatisticsType.QUANTILE, aggregator_type=AggregatorType.MEAN),
    )



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Union

import nncf
from nncf.common.pruning.tensor_processor import NNCFPruningBaseTensorProcessor
from nncf.common.tensor import NNCFTensor


class SymbolicMaskProducer:
    """
    Container of information about a NNCFNode which is produsing a symbolic mask.
    NNCFNode produced a (symbolic or not) mask means this mask was set as an output
    mask to this NNCFNode during (symbolic or not) mask propagation.
    """

    def __init__(self, id_: int, sparse_multiplier: int = 1) -> None:
        self._id = id_
        self._sparse_multiplier = sparse_multiplier

    @property
    def sparse_multiplier(self) -> int:
        return self._sparse_multiplier

    @property
    def id(self) -> int:
        return self._id

    @classmethod
    def merge_producers(cls, masks: List["SymbolicMask"]) -> List["SymbolicMaskProducer"]:
        merged_producers = {}
        for mask in masks:
            for mask_producer in mask.mask_producers:
                if mask_producer.id in merged_producers:
                    assert (
                        mask_producer.sparse_multiplier == merged_producers[mask_producer.id].sparse_multiplier
                    ), f"Inconsistent sparse multiplier for NNCF node with id={mask_producer.id}"
            merged_producers.update({p.id: p for p in mask.mask_producers})
        return list(merged_producers.values())


class SymbolicMask(NNCFTensor):
    """
    Framework agnostic 1D NNCFTensor representation which only uses given dimension and do not uses value
    of the tensor. Keeps additional attribute - symbolic mask producer, pointer to NNCFNode which produced
    this mask during symbolic mask propagation algorithm. NNCFNode produced a (symbolic or not) mask means
    this mask was set as an output mask to this NNCFNode during (symbolic or not) mask propagation.
    Tensor shape and mask producer attributes are correctly propagating during
    symbolic mask propagation by SymbolicMaskProcessor.
    """

    def __init__(self, dimension: int, mask_producers: Union[int, List[SymbolicMaskProducer]] = None):
        super().__init__(None)
        self._mask_producers = mask_producers
        if mask_producers is None:
            self._mask_producers = []
        elif isinstance(mask_producers, int):
            self._mask_producers = [SymbolicMaskProducer(mask_producers)]

        self._shape = dimension

    @property
    def shape(self) -> List[int]:
        return [self._shape]

    @property
    def mask_producers(self) -> List[SymbolicMaskProducer]:
        return self._mask_producers

    @property
    def device(self) -> None:
        return None


class AmbiguousSymbolicMask(SymbolicMask):
    """
    Special case of symbolic mask used when pruning operation
    receive inconsistent set of masks and should produce mask which
    certainly mark all producers of such mask as unprunable by dimension mismatch.
    """

    def __init__(self, mask_producers: Union[int, List[SymbolicMaskProducer]] = None):
        super().__init__(-1, mask_producers)


class SymbolicMaskProcessor(NNCFPruningBaseTensorProcessor):
    """
    Implementation of processing methods set for SymbolicMask.
    Responsible for correct mask dimension and mask producer attributes propagation.
    For methods like concatenate and elementwise_mask_propagation unions
    mask producers of input masks.
    """

    @classmethod
    def concatenate(cls, tensors: List[SymbolicMask], axis: int) -> SymbolicMask:
        ret_shape = sum([t.shape[0] for t in tensors])
        producers = SymbolicMaskProducer.merge_producers(tensors)
        return SymbolicMask(ret_shape, producers)

    @classmethod
    def ones(cls, shape: Union[int, List[int]], device) -> SymbolicMask:
        if isinstance(shape, list):
            if len(shape) != 1:
                raise nncf.ValidationError(f"Unexpected shape = {shape} for 1D symbolic mask")
            shape = shape[0]

        return SymbolicMask(shape)

    @classmethod
    def assert_allclose(cls, tensors: List[SymbolicMask]) -> None:
        for input_mask in tensors[1:]:
            assert tensors[0].shape == input_mask.shape

    @classmethod
    def repeat(cls, tensor: SymbolicMask, repeats: int) -> SymbolicMask:
        updated_mask_producers = []
        for mask_producer in tensor.mask_producers:
            updated_mask_producers.append(
                SymbolicMaskProducer(mask_producer.id, mask_producer.sparse_multiplier * repeats)
            )
        return SymbolicMask(tensor.shape[0] * repeats, updated_mask_producers)

    @classmethod
    def elementwise_mask_propagation(cls, input_masks: List[SymbolicMask]) -> SymbolicMask:
        """
        Assemble output mask for elementwise pruning operation from given input masks.
        In case input_masks have different shape don't propagate any masks.

        :param input_masks: Given input masks.
        :return: Elementwise pruning operation output mask.
        """
        producers = SymbolicMaskProducer.merge_producers(input_masks)
        for input_mask in input_masks[1:]:
            if input_masks[0].shape != input_mask.shape:
                return AmbiguousSymbolicMask(producers)

        return SymbolicMask(input_masks[0].shape[0], producers)

    @classmethod
    def split(cls, tensor: SymbolicMask, output_shapes: List[int]) -> List[SymbolicMask]:
        if any(shape <= 0 for shape in output_shapes) or tensor.shape[0] != sum(output_shapes):
            raise AssertionError(
                "Symbolic mask split was called with"
                f"invalid parammeters: input mask shape: {tensor.shape[0]}, output masks shapes: {output_shapes}"
            )

        producers = tensor.mask_producers
        return [SymbolicMask(output_shape, producers) for output_shape in output_shapes]



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class HWConfigOpName:
    CONVOLUTION = "Convolution"
    DEPTHWISECONVOLUTION = "DepthWiseConvolution"
    MATMUL = "MatMul"
    ADD = "Add"
    MULTIPLY = "Multiply"
    MAXIMUM = "Maximum"
    LESS = "Less"
    LESSEQUAL = "LessEqual"
    GREATER = "Greater"
    GREATEREQUAL = "GreaterEqual"
    DIVIDE = "Divide"
    MINIMUM = "Minimum"
    EQUAL = "Equal"
    SUBTRACT = "Subtract"
    NOTEQUAL = "NotEqual"
    FLOORMOD = "FloorMod"
    LOGICALOR = "LogicalOr"
    LOGICALXOR = "LogicalXor"
    LOGICALAND = "LogicalAnd"
    LOGICALNOT = "LogicalNot"
    POWER = "Power"
    AVGPOOL = "AvgPool"
    REDUCEMEAN = "ReduceMean"
    MAXPOOL = "MaxPool"
    REDUCEMAX = "ReduceMax"
    REDUCESUM = "ReduceSum"
    REDUCEL2 = "ReduceL2"
    INTERPOLATE = "Interpolate"
    MVN = "MVN"
    RESHAPE = "Reshape"
    CONCAT = "Concat"
    FLATTEN = "Flatten"
    SQUEEZE = "Squeeze"
    UNSQUEEZE = "Unsqueeze"
    SPLIT = "Split"
    CHUNK = "Chunk"
    CROP = "Crop"
    TRANSPOSE = "Transpose"
    TILE = "Tile"
    EMBEDDING = "Embedding"
    EMBEDDINGBAG = "EmbeddingBag"
    PAD = "Pad"
    STRIDEDSLICE = "StridedSlice"
    GELU = "Gelu"
    LSTMSEQUENCE = "LSTMSequence"
    GRUSEQUENCE = "GRUSequence"
    GROUPNORMALIZATION = "GroupNormalization"
    SCALED_DOT_PRODUCT_ATTENTION = "ScaledDotProductAttention"



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from copy import deepcopy
from typing import Dict, List

from nncf.common.graph import NNCFNode
from nncf.common.hardware.config import HWConfig
from nncf.common.quantization.structs import QuantizationConstraints
from nncf.common.quantization.structs import QuantizerConfig
from nncf.common.scopes import matches_any


def get_scoped_quantizer_config(
    base_config: QuantizerConfig, scope_str: str, scope_overrides: Dict = None
) -> QuantizerConfig:
    """
    Returns a QuantizerConfig which is based on a given config, which will have overrides
    applied on top of it according to the dictionary of per-scope overrides.

    :param base_config: The base quantizer configuration - corresponding parameters will
      be used in the returned qconfig if no override for this parameter is given.
    :param scope_str: A string identifier of the spot in the model that will be associated with
      this quantizer.
    :param scope_overrides: A dictionary of scope strings vs. dict of overrides for the corresponding
      scope.
    :return: The base configuration with overrides applied on top of it.
    """
    qconfig = deepcopy(base_config)
    if scope_overrides is None:
        scope_overrides = {}
    for overridden_scope in scope_overrides:
        if matches_any(scope_str, overridden_scope):
            config_overrides = scope_overrides[overridden_scope]
            if config_overrides.get("bits") is not None:
                qconfig.num_bits = config_overrides["bits"]
            if config_overrides.get("mode") is not None:
                qconfig.mode = config_overrides["mode"]
            if config_overrides.get("per_channel") is not None:
                qconfig.per_channel = config_overrides["per_channel"]
            if config_overrides.get("signed") is not None:
                qconfig.signedness_to_force = config_overrides["signed"]
    return qconfig


def assign_qconfig_lists_to_modules(
    nodes_with_weights: List[NNCFNode],
    default_weight_qconfig: QuantizerConfig,
    global_weight_constraints: QuantizationConstraints = None,
    scope_overrides_dict: Dict = None,
    hw_config: HWConfig = None,
) -> Dict[NNCFNode, List[QuantizerConfig]]:
    """
    Assigns a list of possible quantizer configurations (as determined by HW config, defaults and overrides)
    to each weighted node that was passed.

    :param nodes_with_weights: The nodes in NNCFGraph that correspond to weighted operations.
    :param default_weight_qconfig: The default quantizer configuration for weights, to be used if
      no other information is given.
    :param global_weight_constraints: The constraints imposed on all weights by the NNCFConfig .json file, such
      as "all symmetric" or "all per-channel" etc.
    :param scope_overrides_dict: The dictionary of strings vs dict of overrides for a given weight quantizer.
      The strings are matched against the name of the NNCFNodes in nodes_with_weights.
    :param hw_config: The HWConfig object to be used for device-specific constraints on allowed weights.
    :return: A dict of each weighted node vs. the list of quantizer configs allowed for quantizing the associated
      weights
    """
    retval: Dict[NNCFNode, List[QuantizerConfig]] = {}
    default_qconfig = deepcopy(default_weight_qconfig)
    if global_weight_constraints is not None:
        default_qconfig = global_weight_constraints.apply_constraints_to(default_qconfig)
    if scope_overrides_dict is None:
        scope_overrides_dict = {}
    weight_scope_overrides_dict = scope_overrides_dict.get("weights")
    if hw_config is not None:
        meta_vs_qconfig_map = hw_config.get_metatype_vs_quantizer_configs_map(for_weights=True)
    for node in nodes_with_weights:
        qconfig_for_current_scope = get_scoped_quantizer_config(
            default_qconfig, node.node_name, weight_scope_overrides_dict
        )
        if hw_config is None:
            qconfig_list = [qconfig_for_current_scope]
        else:
            metatype = node.metatype
            qconfig_list = meta_vs_qconfig_map[metatype]
            if HWConfig.is_wildcard_quantization(qconfig_list):  # Empty list = wildcard quantization
                qconfig_list = [default_qconfig]
            elif HWConfig.is_qconf_list_corresponding_to_unspecified_op(qconfig_list):
                continue  # The module will not have its weights quantized

            local_constraints = global_weight_constraints
            for overridden_scope, scoped_override_dict in scope_overrides_dict.items():
                if matches_any(node.node_name, overridden_scope):
                    scope_constraints = QuantizationConstraints.from_config_dict(scoped_override_dict)
                    local_constraints = local_constraints.get_updated_constraints(scope_constraints)
            qconfig_list = local_constraints.constrain_qconfig_list(
                node.node_name, hw_config.target_device, qconfig_list
            )

        retval[node] = qconfig_list
    return retval



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from types import SimpleNamespace

import numpy as np
import torch
from torch import nn
from torch.autograd import Variable
from torch.optim import Adam

from nncf.torch.automl.agent.ddpg.memory import SequentialMemory

criterion = nn.MSELoss()
USE_CUDA = torch.cuda.is_available()
FLOAT = torch.cuda.FloatTensor if USE_CUDA else torch.FloatTensor


def to_numpy(var):
    return var.cpu().data.numpy() if USE_CUDA else var.data.numpy()


def to_tensor(ndarray, volatile=False, requires_grad=False, dtype=FLOAT):
    return Variable(torch.from_numpy(ndarray), volatile=volatile, requires_grad=requires_grad).type(dtype)


def sample_from_truncated_normal_distribution(lower, upper, mu, sigma, size=1):
    from scipy import stats

    return stats.truncnorm.rvs((lower - mu) / sigma, (upper - mu) / sigma, loc=mu, scale=sigma, size=size)


class Actor(nn.Module):
    def __init__(self, nb_states, nb_actions, hidden1=400, hidden2=300, init_w=3e-3):
        super().__init__()
        self.fc1 = nn.Linear(nb_states, hidden1)
        self.fc2 = nn.Linear(hidden1, hidden2)
        self.fc3 = nn.Linear(hidden2, nb_actions)
        self.relu = nn.ReLU()
        self.sigmoid = nn.Sigmoid()

    def forward(self, x):
        out = self.fc1(x)
        out = self.relu(out)
        out = self.fc2(out)
        out = self.relu(out)
        out = self.fc3(out)
        out = self.sigmoid(out)
        return out


class Critic(nn.Module):
    def __init__(self, nb_states, nb_actions, hidden1=400, hidden2=300, init_w=3e-3):
        super().__init__()
        self.fc11 = nn.Linear(nb_states, hidden1)
        self.fc12 = nn.Linear(nb_actions, hidden1)
        self.fc2 = nn.Linear(hidden1, hidden2)
        self.fc3 = nn.Linear(hidden2, 1)
        self.relu = nn.ReLU()

    def forward(self, xs):
        x, a = xs
        out = self.fc11(x) + self.fc12(a)
        out = self.relu(out)
        out = self.fc2(out)
        out = self.relu(out)
        out = self.fc3(out)
        return out


class DDPG:
    LBOUND = 0.0
    RBOUND = 1.0

    def __init__(self, nb_states, nb_actions, iter_number: int = None, hparam_override: dict = None):
        self.nb_states = nb_states
        self.nb_actions = nb_actions

        hyperparameters = {
            "seed": 528,
            "train_episode": 600,
            "delta_decay": 0.995,
            "hidden1": 300,
            "hidden2": 300,
            "init_w": 0.003,
            "lr_c": 0.001,
            "lr_a": 0.0001,
            "rmsize": 2048,
            "window_length": 1,
            "bsize": 64,
            "tau": 0.01,
            "discount": 1.0,
            "epsilon": 50000,
            "init_delta": 0.5,
            "warmup_iter_number": 20,
            "n_update": 1,
        }

        if hparam_override is not None:
            for hparam, hparam_val in hparam_override.items():
                if hparam in hyperparameters:
                    hyperparameters[hparam] = hparam_val
        if iter_number is not None:
            hyperparameters["train_episode"] = iter_number

        args = SimpleNamespace(**hyperparameters)

        if args.seed is None:
            self.seed(0)
        else:
            self.seed(args.seed)

        # Create Actor and Critic Network
        net_cfg = {"hidden1": args.hidden1, "hidden2": args.hidden2, "init_w": args.init_w}

        self.actor = Actor(self.nb_states, self.nb_actions, **net_cfg)
        self.actor_target = Actor(self.nb_states, self.nb_actions, **net_cfg)
        self.actor_optim = Adam(self.actor.parameters(), lr=args.lr_a)

        self.critic = Critic(self.nb_states, self.nb_actions, **net_cfg)
        self.critic_target = Critic(self.nb_states, self.nb_actions, **net_cfg)
        self.critic_optim = Adam(self.critic.parameters(), lr=args.lr_c)

        self.hard_update(self.actor_target, self.actor)  # Make sure target is with the same weight
        self.hard_update(self.critic_target, self.critic)

        # Create replay buffer
        self.memory = SequentialMemory(limit=args.rmsize, window_length=args.window_length)

        # Hyper-parameters
        self.n_update = args.n_update
        self.batch_size = args.bsize
        self.tau = args.tau
        self.discount = args.discount
        self.depsilon = 1.0 / args.epsilon

        # noise
        self.init_delta = args.init_delta
        self.delta_decay = args.delta_decay
        self.update_delta_decay_factor(args.train_episode)
        self.warmup_iter_number = args.warmup_iter_number
        self.delta = args.init_delta

        # loss
        self.value_loss = 0.0
        self.policy_loss = 0.0

        self.epsilon = 1.0
        self.is_training = True

        if USE_CUDA:
            self.cuda()

        # moving average baseline
        self.moving_average = None
        self.moving_alpha = 0.5  # based on batch, so small

    def update_delta_decay_factor(self, num_train_episode):
        assert num_train_episode > 0, "Number of train episode must be larger than 0"

        if num_train_episode < 1000:
            # every hundred of episode below 1000 (100, 200 ... 1000)
            calibrated_factor = [0.960, 0.960, 0.980, 0.987, 0.992, 0.993, 0.995, 0.995, 0.996, 0.996]
            self.delta_decay = calibrated_factor[num_train_episode // 100]

        elif num_train_episode <= 3000:
            self.delta_decay = (num_train_episode - 999) * (0.999 - 0.997) / (3000 - 999) + 0.997

        else:
            self.delta_decay = 0.999

    def update_policy(self):
        # Sample batch
        state_batch, action_batch, reward_batch, next_state_batch, terminal_batch = self.memory.sample_and_split(
            self.batch_size
        )

        # normalize the reward
        batch_mean_reward = np.mean(reward_batch)
        if self.moving_average is None:
            self.moving_average = batch_mean_reward
        else:
            self.moving_average += self.moving_alpha * (batch_mean_reward - self.moving_average)
        reward_batch -= self.moving_average

        # Prepare for the target q batch
        with torch.no_grad():
            next_q_values = self.critic_target(
                [
                    to_tensor(next_state_batch),
                    self.actor_target(to_tensor(next_state_batch)),
                ]
            )

        target_q_batch = (
            to_tensor(reward_batch) + self.discount * to_tensor(terminal_batch.astype(float)) * next_q_values
        )

        # Critic update
        self.critic.zero_grad()

        q_batch = self.critic([to_tensor(state_batch), to_tensor(action_batch)])

        value_loss = criterion(q_batch, target_q_batch)
        value_loss.backward()
        self.critic_optim.step()

        # Actor update
        self.actor.zero_grad()

        policy_loss = (-1) * self.critic([to_tensor(state_batch), self.actor(to_tensor(state_batch))])

        policy_loss = policy_loss.mean()
        policy_loss.backward()
        self.actor_optim.step()

        # Target update
        self.soft_update(self.actor_target, self.actor)
        self.soft_update(self.critic_target, self.critic)

        # update for log
        self.value_loss = value_loss
        self.policy_loss = policy_loss

    def eval(self):
        self.actor.eval()
        self.actor_target.eval()
        self.critic.eval()
        self.critic_target.eval()

    def cuda(self):
        self.actor.cuda()
        self.actor_target.cuda()
        self.critic.cuda()
        self.critic_target.cuda()

    def observe(self, r_t, s_t, a_t, done):
        if self.is_training:
            self.memory.append(s_t, a_t, r_t, done)  # save to memory

    def random_action(self):
        action = np.random.uniform(self.LBOUND, self.RBOUND, self.nb_actions)
        return action

    def select_action(self, s_t, episode, decay_epsilon=True):
        action = to_numpy(self.actor(to_tensor(np.array(s_t).reshape(1, -1)))).squeeze(0)

        if decay_epsilon is True:
            self.delta = self.init_delta * (self.delta_decay ** (episode - self.warmup_iter_number))
            action = sample_from_truncated_normal_distribution(
                lower=self.LBOUND, upper=self.RBOUND, mu=action, sigma=self.delta, size=self.nb_actions
            )

        return np.clip(action, self.LBOUND, self.RBOUND)

    def reset(self, obs):
        pass

    def load_weights(self, output):
        if output is None:
            return

        self.actor.load_state_dict(torch.load("{}/actor.pkl".format(output)))

        self.critic.load_state_dict(torch.load("{}/critic.pkl".format(output)))

    def save_model(self, output):
        torch.save(self.actor.state_dict(), "{}/actor.pkl".format(output))
        torch.save(self.critic.state_dict(), "{}/critic.pkl".format(output))

    def seed(self, s):
        torch.manual_seed(s)
        if USE_CUDA:
            torch.cuda.manual_seed(s)

    def soft_update(self, target, source):
        for target_param, param in zip(target.parameters(), source.parameters()):
            target_param.data.copy_(target_param.data * (1.0 - self.tau) + param.data * self.tau)

    def hard_update(self, target, source):
        for target_param, param in zip(target.parameters(), source.parameters()):
            target_param.data.copy_(param.data)

    def get_delta(self):
        return self.delta

    def get_value_loss(self):
        return self.value_loss

    def get_policy_loss(self):
        return self.policy_loss



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from copy import deepcopy
from typing import Any, Callable, Iterable, List, Optional, Tuple, TypeVar, Union

import openvino.runtime as ov
from openvino._offline_transformations import compress_quantize_weights_transformation

from nncf.common.factory import NNCFGraphFactory
from nncf.common.logging import nncf_logger
from nncf.common.quantization.structs import QuantizationPreset
from nncf.data import Dataset
from nncf.openvino.graph.metatypes.groups import OPERATIONS_OUTPUT_HAS_NO_BATCH_AXIS
from nncf.openvino.graph.model_utils import remove_friendly_name_duplicates
from nncf.openvino.graph.nncf_graph_builder import GraphConverter
from nncf.openvino.graph.node_utils import get_number_if_op
from nncf.openvino.quantization.backend_parameters import BackendParameters
from nncf.openvino.quantization.backend_parameters import is_weight_compression_needed
from nncf.openvino.quantization.quantize_ifmodel import apply_algorithm_if_bodies
from nncf.openvino.rt_info import dump_parameters
from nncf.parameters import CompressWeightsMode
from nncf.parameters import DropType
from nncf.parameters import ModelType
from nncf.parameters import QuantizationMode
from nncf.parameters import SensitivityMetric
from nncf.parameters import TargetDevice
from nncf.quantization.advanced_parameters import AdvancedAccuracyRestorerParameters
from nncf.quantization.advanced_parameters import AdvancedCompressionParameters
from nncf.quantization.advanced_parameters import AdvancedQuantizationParameters
from nncf.quantization.advanced_parameters import convert_to_dict_recursively
from nncf.quantization.algorithms.accuracy_control.algorithm import QuantizationAccuracyRestorer
from nncf.quantization.algorithms.accuracy_control.algorithm import calculate_accuracy_drop
from nncf.quantization.algorithms.accuracy_control.evaluator import Evaluator
from nncf.quantization.algorithms.post_training.algorithm import PostTrainingQuantization
from nncf.quantization.algorithms.weight_compression.algorithm import WeightCompression
from nncf.quantization.quantize_model import quantize_with_tune_hyperparams
from nncf.quantization.quantize_model import warning_model_no_batchwise_support
from nncf.quantization.telemetry_extractors import CompressionStartedWithQuantizeApi
from nncf.scopes import IgnoredScope
from nncf.telemetry.decorator import tracked_function
from nncf.telemetry.events import NNCF_OV_CATEGORY

TTensor = TypeVar("TTensor")


@tracked_function(NNCF_OV_CATEGORY, [CompressionStartedWithQuantizeApi(), "target_device", "preset"])
def native_quantize_if_op_impl(
    model: ov.Model,
    calibration_dataset: Dataset,
    mode: Optional[QuantizationMode] = None,
    preset: Optional[QuantizationPreset] = None,
    target_device: TargetDevice = TargetDevice.ANY,
    subset_size: int = 300,
    fast_bias_correction: bool = True,
    model_type: Optional[ModelType] = None,
    ignored_scope: Optional[IgnoredScope] = None,
    advanced_parameters: Optional[AdvancedQuantizationParameters] = None,
) -> ov.Model:
    """
    Implementation of the `quantize()` method for the OpenVINO backend via the OpenVINO Runtime API.
    """
    if not fast_bias_correction:
        raise NotImplementedError(
            "The BiasCorrection algorithm is not supported for OpenVINO models with If operation."
        )
    quantization_algorithm = PostTrainingQuantization(
        mode=mode,
        preset=preset,
        target_device=target_device,
        subset_size=subset_size,
        fast_bias_correction=fast_bias_correction,
        model_type=model_type,
        ignored_scope=ignored_scope,
        advanced_parameters=advanced_parameters,
    )

    graph = GraphConverter.create_nncf_graph(model)
    warning_model_no_batchwise_support(graph, advanced_parameters, model_type, OPERATIONS_OUTPUT_HAS_NO_BATCH_AXIS)
    if_ops_number = get_number_if_op(model)
    all_models_number = if_ops_number * 2 + 1
    nncf_logger.info(
        f"The model consists of {if_ops_number} If node(-s) with then and else bodies. \
            Main model and all If bodies will be quantized recursively."
    )
    quantized_model, _ = apply_algorithm_if_bodies(
        quantization_algorithm, model, graph, calibration_dataset, subset_size, 1, all_models_number
    )

    if is_weight_compression_needed(advanced_parameters):
        compress_quantize_weights_transformation(quantized_model)

    dump_parameters(
        quantized_model,
        {
            "preset": preset,
            "target_device": target_device.value,
            "subset_size": subset_size,
            "fast_bias_correction": fast_bias_correction,
            "model_type": model_type,
            "ignored_scope": ignored_scope,
            "advanced_parameters": convert_to_dict_recursively(advanced_parameters),
        },
    )
    return quantized_model


@tracked_function(NNCF_OV_CATEGORY, [CompressionStartedWithQuantizeApi(), "target_device", "preset"])
def native_quantize_impl(
    model: ov.Model,
    calibration_dataset: Dataset,
    mode: Optional[QuantizationMode] = None,
    preset: Optional[QuantizationPreset] = None,
    target_device: TargetDevice = TargetDevice.ANY,
    subset_size: int = 300,
    fast_bias_correction: bool = True,
    model_type: Optional[ModelType] = None,
    ignored_scope: Optional[IgnoredScope] = None,
    advanced_parameters: Optional[AdvancedQuantizationParameters] = None,
) -> ov.Model:
    """
    Implementation of the `quantize()` method for the OpenVINO backend via the OpenVINO Runtime API.
    """
    quantization_algorithm = PostTrainingQuantization(
        mode=mode,
        preset=preset,
        target_device=target_device,
        subset_size=subset_size,
        fast_bias_correction=fast_bias_correction,
        model_type=model_type,
        ignored_scope=ignored_scope,
        advanced_parameters=advanced_parameters,
    )
    graph = GraphConverter.create_nncf_graph(model)
    warning_model_no_batchwise_support(graph, advanced_parameters, model_type, OPERATIONS_OUTPUT_HAS_NO_BATCH_AXIS)
    quantized_model = quantization_algorithm.apply(model, graph, dataset=calibration_dataset)

    if is_weight_compression_needed(advanced_parameters):
        compress_quantize_weights_transformation(quantized_model)

    dump_parameters(
        quantized_model,
        {
            "preset": preset,
            "target_device": target_device.value,
            "subset_size": subset_size,
            "fast_bias_correction": fast_bias_correction,
            "model_type": model_type,
            "ignored_scope": ignored_scope,
            "advanced_parameters": convert_to_dict_recursively(advanced_parameters),
        },
    )
    return quantized_model


@tracked_function(
    NNCF_OV_CATEGORY, [CompressionStartedWithQuantizeApi(), "target_device", "preset", "max_drop", "drop_type"]
)
def native_quantize_with_accuracy_control_impl(
    model: ov.Model,
    calibration_dataset: Dataset,
    validation_dataset: Dataset,
    validation_fn: Callable[[Any, Iterable[Any]], Tuple[float, Union[None, List[float], List[List[TTensor]]]]],
    max_drop: float = 0.01,
    drop_type: DropType = DropType.ABSOLUTE,
    preset: Optional[QuantizationPreset] = None,
    target_device: TargetDevice = TargetDevice.ANY,
    subset_size: int = 300,
    fast_bias_correction: bool = True,
    model_type: Optional[ModelType] = None,
    ignored_scope: Optional[IgnoredScope] = None,
    advanced_quantization_parameters: Optional[AdvancedQuantizationParameters] = None,
    advanced_accuracy_restorer_parameters: Optional[AdvancedAccuracyRestorerParameters] = None,
) -> ov.Model:
    """
    Implementation of the `quantize_with_accuracy_control()` method for the OpenVINO backend via the
    OpenVINO Runtime API.
    """
    if advanced_accuracy_restorer_parameters is None:
        advanced_accuracy_restorer_parameters = AdvancedAccuracyRestorerParameters()

    compress_weights = is_weight_compression_needed(advanced_quantization_parameters)

    if advanced_quantization_parameters is None:
        copied_parameters = AdvancedQuantizationParameters()
    else:
        copied_parameters = deepcopy(advanced_quantization_parameters)
    copied_parameters.backend_params[BackendParameters.COMPRESS_WEIGHTS] = False

    quantized_model = quantize_impl(
        model=model,
        calibration_dataset=calibration_dataset,
        preset=preset,
        target_device=target_device,
        subset_size=subset_size,
        fast_bias_correction=fast_bias_correction,
        model_type=model_type,
        ignored_scope=ignored_scope,
        advanced_parameters=copied_parameters,
    )

    if advanced_accuracy_restorer_parameters.intermediate_model_dir:
        quantized_model_path = f"{advanced_accuracy_restorer_parameters.intermediate_model_dir}/intermediate_model.xml"
        ov.serialize(quantized_model, quantized_model_path)

    evaluator = Evaluator(validation_fn)
    evaluator.enable_iteration_count()
    initial_metric_results = evaluator.collect_metric_results(model, validation_dataset, model_name="initial")
    validation_dataset_size = evaluator.num_passed_iterations
    evaluator.disable_iteration_count()

    quantized_metric_results = evaluator.collect_metric_results(
        quantized_model, validation_dataset, model_name="quantized"
    )

    should_terminate, accuracy_drop = calculate_accuracy_drop(
        initial_metric_results.metric_value, quantized_metric_results.metric_value, max_drop, drop_type
    )

    nncf_logger.info(f"Accuracy drop: {accuracy_drop} ({drop_type})")

    # TODO(andrey-churkin): Collect statistics only once
    if advanced_accuracy_restorer_parameters.tune_hyperparams and not should_terminate:
        model = remove_friendly_name_duplicates(model)
        tuned_quantized_model = quantize_with_tune_hyperparams(
            model,
            calibration_dataset,
            validation_dataset,
            validation_fn,
            initial_metric_results,
            quantized_metric_results,
            subset_size,
            preset,
            target_device,
            subset_size,
            fast_bias_correction,
            model_type,
            ignored_scope,
            copied_parameters,
        )
        tuned_quantized_metric_results = evaluator.collect_metric_results(
            tuned_quantized_model, validation_dataset, model_name="tuned"
        )
        should_terminate, tuned_accuracy_drop = calculate_accuracy_drop(
            initial_metric_results.metric_value, tuned_quantized_metric_results.metric_value, max_drop, drop_type
        )

        nncf_logger.info(f"Accuracy drop (tuned): {tuned_accuracy_drop} ({drop_type})")

        if should_terminate or tuned_accuracy_drop < accuracy_drop:
            quantized_model = tuned_quantized_model
            quantized_metric_results = tuned_quantized_metric_results

    if not should_terminate:
        ranking_subset_size = subset_size
        if advanced_accuracy_restorer_parameters.ranking_subset_size is not None:
            ranking_subset_size = advanced_accuracy_restorer_parameters.ranking_subset_size

        accuracy_restorer = QuantizationAccuracyRestorer(
            ranking_subset_size,
            advanced_accuracy_restorer_parameters.max_num_iterations,
            max_drop,
            drop_type,
            advanced_accuracy_restorer_parameters.num_ranking_workers,
            advanced_accuracy_restorer_parameters.restore_mode,
        )
        quantized_model = accuracy_restorer.apply(
            model,
            initial_metric_results,
            quantized_model,
            quantized_metric_results,
            validation_dataset,
            validation_dataset_size,
            evaluator,
        )

    if compress_weights:
        compress_quantize_weights_transformation(quantized_model)

    dump_parameters(
        quantized_model,
        {
            "preset": preset,
            "target_device": target_device.value,
            "subset_size": subset_size,
            "fast_bias_correction": fast_bias_correction,
            "model_type": model_type,
            "ignored_scope": ignored_scope,
            "max_drop": max_drop,
            "drop_type": drop_type.value,
            "advanced_quantization_parameters": convert_to_dict_recursively(advanced_quantization_parameters),
            "advanced_accuracy_restorer_parameters": convert_to_dict_recursively(advanced_accuracy_restorer_parameters),
        },
    )
    return quantized_model


def quantize_impl(
    model: ov.Model,
    calibration_dataset: Dataset,
    mode: Optional[QuantizationMode] = None,
    preset: Optional[QuantizationPreset] = None,
    target_device: TargetDevice = TargetDevice.ANY,
    subset_size: int = 300,
    fast_bias_correction: bool = True,
    model_type: Optional[ModelType] = None,
    ignored_scope: Optional[IgnoredScope] = None,
    advanced_parameters: Optional[AdvancedQuantizationParameters] = None,
) -> ov.Model:
    """
    Implementation of the `quantize()` method for the OpenVINO backend.
    """
    model = remove_friendly_name_duplicates(model)

    quantize_fn = native_quantize_impl
    if get_number_if_op(model) > 0:
        quantize_fn = native_quantize_if_op_impl

    return quantize_fn(
        model=model,
        calibration_dataset=calibration_dataset,
        mode=mode,
        preset=preset,
        target_device=target_device,
        subset_size=subset_size,
        fast_bias_correction=fast_bias_correction,
        model_type=model_type,
        ignored_scope=ignored_scope,
        advanced_parameters=advanced_parameters,
    )


def wrap_validation_fn(validation_fn):
    """
    Wraps validation function to support case when it only returns metric value.

    :param validation_fn: Validation function to wrap.
    :return: Wrapped validation function.
    """

    def wrapper(*args, **kwargs):
        retval = validation_fn(*args, **kwargs)
        if isinstance(retval, tuple):
            return retval
        return retval, None

    return wrapper


def quantize_with_accuracy_control_impl(
    model: ov.Model,
    calibration_dataset: Dataset,
    validation_dataset: Dataset,
    validation_fn: Callable[[Any, Iterable[Any]], float],
    max_drop: float = 0.01,
    drop_type: DropType = DropType.ABSOLUTE,
    preset: Optional[QuantizationPreset] = None,
    target_device: TargetDevice = TargetDevice.ANY,
    subset_size: int = 300,
    fast_bias_correction: bool = True,
    model_type: Optional[ModelType] = None,
    ignored_scope: Optional[IgnoredScope] = None,
    advanced_quantization_parameters: Optional[AdvancedQuantizationParameters] = None,
    advanced_accuracy_restorer_parameters: Optional[AdvancedAccuracyRestorerParameters] = None,
) -> ov.Model:
    """
    Implementation of the `quantize_with_accuracy_control()` method for the OpenVINO backend.
    """

    quantize_with_accuracy_control_fn = native_quantize_with_accuracy_control_impl

    val_func = wrap_validation_fn(validation_fn)

    return quantize_with_accuracy_control_fn(
        model,
        calibration_dataset,
        validation_dataset,
        val_func,
        max_drop,
        drop_type,
        preset,
        target_device,
        subset_size,
        fast_bias_correction,
        model_type,
        ignored_scope,
        advanced_quantization_parameters,
        advanced_accuracy_restorer_parameters,
    )


def compress_weights_impl(
    model: ov.Model,
    dataset: Dataset,
    mode: CompressWeightsMode,
    ratio: float,
    group_size: int,
    ignored_scope: IgnoredScope,
    all_layers: bool,
    sensitivity_metric: SensitivityMetric,
    awq: bool,
    subset_size: int,
    scale_estimation: bool,
    gptq: bool,
    advanced_parameters: Optional[AdvancedCompressionParameters] = None,
) -> ov.Model:
    """
    Implementation of the `compress_weights()` method for the OpenVINO backend.
    """

    model = remove_friendly_name_duplicates(model)
    compression_algorithm = WeightCompression(
        mode,
        ratio,
        group_size,
        ignored_scope,
        all_layers,
        sensitivity_metric,
        awq,
        subset_size,
        scale_estimation,
        gptq,
        advanced_parameters,
    )
    graph = NNCFGraphFactory.create(model)
    return compression_algorithm.apply(model, graph, dataset=dataset)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import nncf


def get_built_model(model, config):
    if not model.built:
        input_info = config.get("input_info", {})
        if isinstance(input_info, dict):
            sample_size = input_info.get("sample_size", None)
        else:
            sample_size = input_info[0].get("sample_size", None) if input_info else None
        if not sample_size:
            raise nncf.ValidationError("sample_size must be provided in configuration file")
        model.build([None] + list(sample_size[1:]))

    return model



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
from copy import deepcopy
from enum import Enum
from typing import Dict, List, Optional, Set

import networkx as nx

from nncf.common.graph import NNCFGraph
from nncf.common.graph import NNCFNodeName
from nncf.common.graph.graph import NNCFNode
from nncf.common.graph.graph_matching import find_subgraphs_matching_pattern
from nncf.common.graph.layer_attributes import Dtype
from nncf.common.graph.operator_metatypes import INPUT_NOOP_METATYPES
from nncf.common.graph.patterns import GraphPattern
from nncf.common.logging import nncf_logger


class InsertionPointGraphNodeType(Enum):
    PRE_HOOK = 0
    POST_HOOK = 1
    OPERATOR = 2


class PreHookInsertionPoint:
    def __init__(self, target_node_name: str, input_port_id: int):
        self.target_node_name = target_node_name
        self.input_port_id = input_port_id

    def __str__(self):
        return str(self.input_port_id) + " " + self.target_node_name


class PostHookInsertionPoint:
    def __init__(self, target_node_name: str):
        self.target_node_name = target_node_name

    def __str__(self):
        return self.target_node_name


class InsertionPointGraph(nx.DiGraph):
    """
    This graph is built from the NNCFGraph representation of the model control flow graph and adds ephemeral
    "insertion point nodes" into the NNCF model graph representation corresponding to operator pre- and
    post-hooks. Module pre-op and post-op insertion points are currently not reflected here, but they are
    probably not required for quantizing activations, for which the quantizer propagation makes sense.
    This "insertion point graph" representation is useful for quantizer propagation and for referencing
    the compression algorithm hooks to the model operations to which they are applied to.
    """

    NODE_TYPE_NODE_ATTR = "node_type"
    INSERTION_POINT_NODE_ATTR = "insertion_point"
    IS_IN_NNCF_MODULE_NODE_ATTR = "is_in_nncf_module"
    REGULAR_NODE_REF_NODE_ATTR = "regular_node_data"
    ASSOCIATED_IP_NODE_KEYS_NODE_ATTR = "associated_ip_node_keys"
    IS_MERGED_NODE_ATTR = "is_merged"
    MERGED_NNCF_NODE_LIST_NODE_ATTR = "merged_node_list"
    IS_INTEGER_PATH_EDGE_ATTR = "is_integer"

    PRE_HOOK_ID_PREFIX = "PRE HOOK "  # NB: Do not use colon (':') in node keys! Causes trouble for .dot file export.
    POST_HOOK_ID_PREFIX = "POST HOOK "

    def __init__(
        self,
        nncf_graph: NNCFGraph,
        weight_modifiable_node_names: List[NNCFNodeName] = None,
        allowed_pre_hook_insertion_points: List[PreHookInsertionPoint] = None,
        allowed_post_hook_insertion_points: List[PostHookInsertionPoint] = None,
    ):
        """
        Initializes the insertion point graph.

        :param nncf_graph: The base NNCFGraph representing the model structure.
        :param weight_modifiable_node_names: Names of the nodes in `nncf_graph` that correspond to operations with
          modifiable weights.
        :param allowed_pre_hook_insertion_points: A list of pre-hook insertion points for this graph to allow.
          If left unspecified, every node in `nncf_graph` will be allowed to have a separate pre-hook for each of its
          tensor inputs.
        :param allowed_post_hook_insertion_points: A list of post-hook insertion points for this graph to allow.
        If left unspecified, every node in `nncf_graph` will be allowed to have a single post-hook for its output
         (post-hooking separate tensors in an operation's output is not currently supported)
        """

        super().__init__()
        self._base_nx_graph = deepcopy(nncf_graph.get_nx_graph_copy())
        if weight_modifiable_node_names is None:
            self._weight_modifiable_node_names = []
        else:
            self._weight_modifiable_node_names = weight_modifiable_node_names

        if allowed_pre_hook_insertion_points is None:
            allowed_pre_hook_insertion_points = self._get_default_pre_hook_ip_list(nncf_graph)

        if allowed_post_hook_insertion_points is None:
            # Post-hook all nodes if an exact list is not specified
            allowed_post_hook_insertion_points = self._get_default_post_hook_ip_list(nncf_graph)

        target_node_name_vs_pre_hook_ips: Dict[NNCFNodeName, Set[PreHookInsertionPoint]] = defaultdict(set)
        for pre_hook_ip in allowed_pre_hook_insertion_points:
            target_node_name_vs_pre_hook_ips[pre_hook_ip.target_node_name].add(pre_hook_ip)

        target_node_name_vs_post_hook_ips: Dict[NNCFNodeName, Set[PostHookInsertionPoint]] = defaultdict(set)
        for post_hook_ip in allowed_post_hook_insertion_points:
            target_node_name_vs_post_hook_ips[post_hook_ip.target_node_name].add(post_hook_ip)

        for node_key in nx.lexicographical_topological_sort(self._base_nx_graph):
            nncf_node = nncf_graph.get_node_by_key(node_key)
            attrs = {
                InsertionPointGraph.REGULAR_NODE_REF_NODE_ATTR: nncf_node,
                InsertionPointGraph.NODE_TYPE_NODE_ATTR: InsertionPointGraphNodeType.OPERATOR,
                InsertionPointGraph.ASSOCIATED_IP_NODE_KEYS_NODE_ATTR: set(),
                InsertionPointGraph.IS_MERGED_NODE_ATTR: False,
            }
            self.add_node(node_key, **attrs)

        INPUT_PORT_ID = "input_port_id"
        for edge in self._base_nx_graph.edges:
            input_port_id = self._base_nx_graph.edges[edge][NNCFGraph.INPUT_PORT_ID_EDGE_ATTR]
            dtype = self._base_nx_graph.edges[edge][NNCFGraph.DTYPE_EDGE_ATTR]
            parallel_input_port_ids = self._base_nx_graph.edges[edge][NNCFGraph.PARALLEL_INPUT_PORT_IDS_ATTR]
            from_node, to_node = edge

            attrs = {
                INPUT_PORT_ID: input_port_id,
                self.IS_INTEGER_PATH_EDGE_ATTR: dtype is Dtype.INTEGER,
                NNCFGraph.PARALLEL_INPUT_PORT_IDS_ATTR: parallel_input_port_ids,
            }
            self.add_edge(from_node, to_node, **attrs)

        node_keys_working_set = [deepcopy(node_key) for node_key in nx.lexicographical_topological_sort(self)]

        # TODO (vshampor): Add insertion points for module pre- and post-ops.
        # Should roughly look so: first, determine subsets of nodes belonging to each
        # separate NNCF module (via scope analysis), then for each subset find input/output
        # edges using a corresponding NNCFGraph function; add a pre-op insertion point node as the
        # sink for input edges and connect it to input edge destinations, then add a post-op
        # insertion point as the source of output edges and connect it to output edge origins.

        for operator_node_key in node_keys_working_set:
            operator_node = self.nodes[operator_node_key]
            original_node = operator_node[InsertionPointGraph.REGULAR_NODE_REF_NODE_ATTR]
            if original_node.node_name in target_node_name_vs_pre_hook_ips:
                pre_hook_ips = list(target_node_name_vs_pre_hook_ips[original_node.node_name])
                pre_hook_ips = sorted(pre_hook_ips, key=lambda x: x.input_port_id)
                in_edges = list(self.in_edges(operator_node_key))
                input_port_id_vs_edge = {}
                for edge in in_edges:
                    input_port_id = self.edges[edge][INPUT_PORT_ID]
                    input_port_id_vs_edge[input_port_id] = edge
                    for parallel_input_port_id in self.edges[edge][NNCFGraph.PARALLEL_INPUT_PORT_IDS_ATTR]:
                        input_port_id_vs_edge[parallel_input_port_id] = edge

                encountered_input_edges = set()
                for pre_hook_point in pre_hook_ips:
                    edge = input_port_id_vs_edge[pre_hook_point.input_port_id]
                    original_edge_attrs = self.edges[edge]
                    from_node_key, to_node_key = edge
                    ip_node_key = self.get_pre_hook_node_key(str(operator_node_key), pre_hook_point.input_port_id)

                    pre_hook_ip_attrs = {
                        InsertionPointGraph.NODE_TYPE_NODE_ATTR: InsertionPointGraphNodeType.PRE_HOOK,
                        InsertionPointGraph.INSERTION_POINT_NODE_ATTR: pre_hook_point,
                    }

                    self.add_node(ip_node_key, **pre_hook_ip_attrs)

                    encountered_input_edges.add(edge)
                    self.add_edge(from_node_key, ip_node_key, **original_edge_attrs)
                    self.add_edge(ip_node_key, operator_node_key, **original_edge_attrs)
                    operator_node[InsertionPointGraph.ASSOCIATED_IP_NODE_KEYS_NODE_ATTR].add(ip_node_key)

                for edge in encountered_input_edges:
                    self.remove_edge(*edge)

            if original_node.node_name in target_node_name_vs_post_hook_ips:
                post_hook_ips = target_node_name_vs_post_hook_ips[original_node.node_name]
                assert len(post_hook_ips) == 1, "Multiple post-hooks for a single NNCFGraph node are not supported!"
                post_hook_ip = next(iter(post_hook_ips))
                post_hook_ip_attrs = {
                    InsertionPointGraph.NODE_TYPE_NODE_ATTR: InsertionPointGraphNodeType.POST_HOOK,
                    InsertionPointGraph.INSERTION_POINT_NODE_ATTR: post_hook_ip,
                }
                ip_node_key = self.get_post_hook_node_key(str(operator_node_key))
                self.add_node(ip_node_key, **post_hook_ip_attrs)
                out_edges = list(self.out_edges(operator_node_key))
                has_integer_outputs = False
                for out_edge in out_edges:
                    # Need to preserve original edge attributes in order not to lose
                    # input port ID information
                    original_edge_attrs = self.edges[out_edge]
                    from_node_key, to_node_key = out_edge
                    self.remove_edge(from_node_key, to_node_key)
                    self.add_edge(ip_node_key, to_node_key, **original_edge_attrs)
                    if original_edge_attrs[self.IS_INTEGER_PATH_EDGE_ATTR]:
                        has_integer_outputs = True

                    # TODO (vshampor): introduce separate insertion points for operator outputs if
                    # the outputs are semantically different

                # TODO (vshampor): in multi-output case, some outputs may be integer and some float;
                #  need to switch to using output ports to cover this correctly. For safety, mark
                #  the edge from op to post-hook as integer if at least one output edge of the op was integer
                is_integer_attrs = {InsertionPointGraph.IS_INTEGER_PATH_EDGE_ATTR: has_integer_outputs}
                self.add_edge(operator_node_key, ip_node_key, **is_integer_attrs)
                operator_node = self.nodes[operator_node_key]
                operator_node[InsertionPointGraph.ASSOCIATED_IP_NODE_KEYS_NODE_ATTR].add(ip_node_key)

        for edge in self.edges:
            # Mark all edges from post-hook to pre-hook as integer if at least one was integer.
            # Until output ports are ready, the post-hook for output will treat op as having a single
            # tensor output. In multi-output case when some of tensors are integer, need to make
            # sure that the propagation won't happen from a pre-hook of the op consuming the floating part
            # of the output into the post-hook of the operation that produces both int and float tensors.
            from_node_key, to_node_key = edge
            from_node = self.nodes[from_node_key]
            to_node = self.nodes[to_node_key]
            if (
                from_node[self.NODE_TYPE_NODE_ATTR] is InsertionPointGraphNodeType.POST_HOOK
                and to_node[self.NODE_TYPE_NODE_ATTR] is InsertionPointGraphNodeType.PRE_HOOK
            ):
                post_hook_has_integer_outputs = False
                for follower_node_key in self.successors(from_node_key):
                    if self.edges[from_node_key, follower_node_key][self.IS_INTEGER_PATH_EDGE_ATTR]:
                        post_hook_has_integer_outputs = True
                if post_hook_has_integer_outputs:
                    for follower_node_key in self.successors(from_node_key):
                        self.edges[from_node_key, follower_node_key][self.IS_INTEGER_PATH_EDGE_ATTR] = True

    @property
    def weight_modifiable_node_names(self) -> List[NNCFNodeName]:
        return self._weight_modifiable_node_names

    @staticmethod
    def _get_default_pre_hook_ip_list(nncf_graph: NNCFGraph) -> List[PreHookInsertionPoint]:
        # Pre-hook all input ports of all nodes
        allowed_pre_hook_insertion_points = []
        for nncf_node in nncf_graph.get_all_nodes():
            pred_nodes = nncf_graph.get_previous_nodes(nncf_node)

            for pred_node in pred_nodes:
                input_edge = nncf_graph.get_edge(pred_node, nncf_node)
                allowed_pre_hook_insertion_points.append(
                    PreHookInsertionPoint(nncf_node.node_name, input_edge.input_port_id)
                )
        return allowed_pre_hook_insertion_points

    @staticmethod
    def _get_default_post_hook_ip_list(nncf_graph: NNCFGraph) -> List[PostHookInsertionPoint]:
        # Post-hook all nodes, post hook applies to the entire op output
        allowed_post_hook_insertion_points = []
        for nncf_node in nncf_graph.get_all_nodes():
            allowed_post_hook_insertion_points.append(PostHookInsertionPoint(nncf_node.node_name))
        return allowed_post_hook_insertion_points

    def remove_nodes_from(self, nodes: List[str]) -> None:
        """
        Removes nodes from the InsertionPointGraph and its _base_nx_graph.

        :param nodes: List of the nodes to remove.
        :return:
        """
        super().remove_nodes_from(nodes)
        constant_nodes_base_nx_graph = [node for node in nodes if node in self._base_nx_graph]
        self._base_nx_graph.remove_nodes_from(constant_nodes_base_nx_graph)

    def get_base_nx_graph(self) -> nx.DiGraph:
        """
        Returns the self._base_nx_graph.

        :return: The self._base_nx_graph.
        """
        return self._base_nx_graph

    def get_input_nodes(self) -> List[str]:
        """
        Returns all input nodes, meaning the nodes which belong to any of INPUT_NOOP_METATYPES metatype.

        :return: A list of input nodes.
        """
        output = []
        for node, data in self.nodes.items():
            if InsertionPointGraph.REGULAR_NODE_REF_NODE_ATTR not in data:
                continue
            if data[InsertionPointGraph.IS_MERGED_NODE_ATTR]:
                for nncf_node in data[InsertionPointGraph.MERGED_NNCF_NODE_LIST_NODE_ATTR]:
                    if self._base_nx_graph.nodes[nncf_node.node_key][NNCFNode.METATYPE_ATTR] in INPUT_NOOP_METATYPES:
                        output.append(node)
                        break
            elif data[InsertionPointGraph.REGULAR_NODE_REF_NODE_ATTR].metatype in INPUT_NOOP_METATYPES:
                output.append(node)
        return output

    def get_merged_node_from_single_node_key(self, node_key: str) -> str:
        """
        Returns the node key of the composite node,
        if the corresponding node with 'node_key' was merged into it during the graph initialization.
        Otherwise, returns the original 'node_key'.

        :param node_key: The key of the node which is checking on the merged.
        :return: The node key of the composite node. Original 'node_key' if the node was not merged.
        """
        for node, data in self.nodes.items():
            if InsertionPointGraph.REGULAR_NODE_REF_NODE_ATTR not in data:
                continue
            if data[InsertionPointGraph.IS_MERGED_NODE_ATTR]:
                for nncf_node in data[InsertionPointGraph.MERGED_NNCF_NODE_LIST_NODE_ATTR]:
                    if node_key == nncf_node.node_key:
                        return node
        return node_key

    def get_ip_graph_with_merged_hw_optimized_operations(
        self, full_fusing_pattern: GraphPattern
    ) -> "InsertionPointGraph":
        """
        Returns an InsertionPointGraph in which the nodes that match a HW-specific list of patterns are fused into a
        single node; the resulting InsertionPointGraph no longer has accessible the pre- and post-hooks that were
        located in  the middle of the fused pattern.
        If the InsertionPointGraph should be filtered from constant nodes before the node fusing,
        then 'known_non_constant_node_keys' should be pass. This is the list of the node known that are non constansts.

        :param full_fusing_pattern: The GraphPatttern object representing a composition of fusing pattern variants.
        :return: The InsertionPointGraph with nodes fused according to pattern matching.
        """

        merged_ip_graph = deepcopy(self)
        matches = find_subgraphs_matching_pattern(merged_ip_graph.get_base_nx_graph(), full_fusing_pattern)
        for match in matches:
            if len(match) == 1:
                continue

            input_node_key = match[0]
            output_node_key = match[-1]

            in_edges = list(self.in_edges(input_node_key))
            out_edges = list(self.out_edges(output_node_key))

            in_edge_copies_dict = {}
            for in_edge_key in in_edges:
                in_edge_copies_dict[in_edge_key] = deepcopy(self.edges[in_edge_key])
            out_edge_copies_dict = {}
            for out_edge_key in out_edges:
                out_edge_copies_dict[out_edge_key] = deepcopy(self.edges[out_edge_key])

            conserved_edges_list = out_edges + in_edges

            merged_node_attrs = deepcopy(self.nodes[input_node_key])
            merged_node_attrs[InsertionPointGraph.ASSOCIATED_IP_NODE_KEYS_NODE_ATTR] = set()
            merged_node_attrs[InsertionPointGraph.IS_MERGED_NODE_ATTR] = True
            merged_node_key = ""
            merged_nncf_nodes = []
            for node_key in match:
                ip_node_keys = self.nodes[node_key][InsertionPointGraph.ASSOCIATED_IP_NODE_KEYS_NODE_ATTR]
                for ip_node_key in ip_node_keys:
                    should_keep_ip_node = False
                    for edge_key in conserved_edges_list:
                        if ip_node_key in edge_key:
                            should_keep_ip_node = True
                            break
                    if should_keep_ip_node:
                        merged_node_attrs[InsertionPointGraph.ASSOCIATED_IP_NODE_KEYS_NODE_ATTR].add(ip_node_key)
                    else:
                        merged_ip_graph.remove_node(ip_node_key)
                merged_nncf_nodes.append(self.nodes[node_key][InsertionPointGraph.REGULAR_NODE_REF_NODE_ATTR])
                merged_ip_graph.remove_node(node_key)
                merged_node_key += node_key + "\n"

            # The first node in the merged node list will be considered a "primary" node for purposes
            # of further ignored/target scope application.
            merged_node_attrs[InsertionPointGraph.MERGED_NNCF_NODE_LIST_NODE_ATTR] = merged_nncf_nodes
            merged_ip_graph.add_node(merged_node_key, **merged_node_attrs)
            for in_edge_key, in_edge_attrs in in_edge_copies_dict.items():
                merged_ip_graph.add_edge(in_edge_key[0], merged_node_key, **in_edge_attrs)
            for out_edge_key, out_edge_attrs in out_edge_copies_dict.items():
                merged_ip_graph.add_edge(merged_node_key, out_edge_key[1], **out_edge_attrs)

        return merged_ip_graph

    @staticmethod
    def get_pre_hook_node_key(node_key: str, input_port_id: int = 0) -> str:
        return InsertionPointGraph.PRE_HOOK_ID_PREFIX + str(input_port_id) + " " + node_key

    @staticmethod
    def get_post_hook_node_key(node_key: str) -> str:
        return InsertionPointGraph.POST_HOOK_ID_PREFIX + node_key


class ConstantNodesFilter:
    @staticmethod
    def filter(ip_graph: InsertionPointGraph, start_traversing_node_keys: Optional[List[str]]) -> InsertionPointGraph:
        """
        Removes all Constant nodes from InsertionPointGraph, making it inference graph.
        The traversing starts from the input nodes and nodes with weights.

        :param ip_graph: The original InsertionPointGraph.
        :param start_traversing_node_keys: Keys of the nodes from which the traversing will be start.
        :return: InsertionPointGraph without Constant nodes.
        """
        input_nodes = ip_graph.get_input_nodes()
        if not input_nodes:
            nncf_logger.debug("Skipped filtering - no input nodes found")
            return ip_graph
        weight_nodes = []
        if start_traversing_node_keys is not None:
            weight_nodes = [
                ip_graph.get_merged_node_from_single_node_key(weight_node) for weight_node in start_traversing_node_keys
            ]
        visited_nodes = set()
        start_nodes = input_nodes + weight_nodes
        for node in start_nodes:
            for node_from, node_to in nx.bfs_edges(ip_graph, source=node):
                visited_nodes.add(node_from)
                visited_nodes.add(node_to)
        constant_nodes = [node for node in ip_graph.nodes if node not in visited_nodes]
        ip_graph.remove_nodes_from(constant_nodes)
        return ip_graph



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Tuple

import numpy as np

from nncf.common.graph import NNCFGraph
from nncf.common.graph import NNCFNodeName
from nncf.common.graph.operator_metatypes import OperatorMetatype
from nncf.common.pruning.utils import get_output_channels
from nncf.common.pruning.utils import is_prunable_depthwise_conv


class WeightsFlopsCalculator:
    """
    Collection of weight and flops calculation functions.
    Class instance keeps only parameters that are constant during
    compression algorithms execution.
    """

    def __init__(self, conv_op_metatypes: List[OperatorMetatype], linear_op_metatypes: List[OperatorMetatype]):
        """
        Constructor.

        :param conv_op_metatypes: List of metatypes defining convolution operations.
        :param linear_op_metatypes: List of metatypes defining linear/fully connected operations.
        """
        self._conv_op_metatypes = conv_op_metatypes
        self._linear_op_metatypes = linear_op_metatypes

    def count_flops_and_weights(
        self,
        graph: NNCFGraph,
        output_shapes: Dict[NNCFNodeName, int],
        input_channels: Dict[NNCFNodeName, int] = None,
        output_channels: Dict[NNCFNodeName, int] = None,
        kernel_sizes: Dict[NNCFNodeName, Tuple[int, int]] = None,
        op_addresses_to_skip: List[str] = None,
    ) -> Tuple[int, int]:
        """
        Counts the number of weights and FLOPs in the model for convolution and fully connected layers.

        :param graph: NNCFGraph.
        :param output_shapes: Dictionary of output dimension shapes for convolutions and
            fully connected layers. E.g {node_name: (height, width)}
        :param input_channels: Dictionary of input channels number in convolutions.
            If not specified, taken from the graph. {node_name: channels_num}
        :param output_channels: Dictionary of output channels number in convolutions.
            If not specified, taken from the graph. {node_name: channels_num}
        :param kernel_sizes: Dictionary of kernel sizes in convolutions.
            If not specified, taken from the graph. {node_name: kernel_size}.
            It's only supposed to be used in NAS in case of Elastic Kernel enabled.
        :param op_addresses_to_skip: List of operation addresses of layers that should be skipped from calculation.
            It's only supposed to be used in NAS in case of Elastic Depth enabled.
        :return number of FLOPs for the model
                number of weights (params) in the model
        """
        flops_pers_node, weights_per_node = self.count_flops_and_weights_per_node(
            graph, output_shapes, input_channels, output_channels, kernel_sizes, op_addresses_to_skip
        )
        return sum(flops_pers_node.values()), sum(weights_per_node.values())

    def count_flops_and_weights_per_node(
        self,
        graph: NNCFGraph,
        output_shapes: Dict[NNCFNodeName, int],
        input_channels: Dict[NNCFNodeName, int] = None,
        output_channels: Dict[NNCFNodeName, int] = None,
        kernel_sizes: Dict[NNCFNodeName, Tuple[int, int]] = None,
        op_addresses_to_skip: List[NNCFNodeName] = None,
    ) -> Tuple[Dict[NNCFNodeName, int], Dict[NNCFNodeName, int]]:
        """
        Counts the number of weights and FLOPs per node in the model for convolution and fully connected layers.

        :param graph: NNCFGraph.
        :param output_shapes: Dictionary of output dimension shapes for convolutions and
            fully connected layers. E.g {node_name: (height, width)}
        :param input_channels: Dictionary of input channels number in convolutions.
            If not specified, taken from the graph. {node_name: channels_num}
        :param output_channels: Dictionary of output channels number in convolutions.
            If not specified, taken from the graph. {node_name: channels_num}
        :param kernel_sizes: Dictionary of kernel sizes in convolutions.
            If not specified, taken from the graph. {node_name: kernel_size}.
            It's only supposed to be used in NAS in case of Elastic Kernel enabled.
        :param op_addresses_to_skip: List of operation addresses of layers that should be skipped from calculation.
            It's only supposed to be used in NAS in case of Elastic Depth enabled.
        :return Dictionary of FLOPs number {node_name: flops_num}
                Dictionary of weights number {node_name: weights_num}
        """
        flops = {}
        weights = {}
        input_channels = input_channels or {}
        output_channels = output_channels or {}
        kernel_sizes = kernel_sizes or {}
        op_addresses_to_skip = op_addresses_to_skip or []
        for node in graph.get_nodes_by_metatypes(self._conv_op_metatypes):
            name = node.node_name
            if name in op_addresses_to_skip:
                continue
            num_in_channels = input_channels.get(name, node.layer_attributes.in_channels)
            num_out_channels = output_channels.get(name, node.layer_attributes.out_channels)
            kernel_size = kernel_sizes.get(name, node.layer_attributes.kernel_size)
            if is_prunable_depthwise_conv(node):
                # Prunable depthwise conv processed in special way
                # because common way to calculate filters per
                # channel for such layer leads to zero in case
                # some of the output channels are pruned.
                filters_per_channel = 1
            else:
                filters_per_channel = num_out_channels // node.layer_attributes.groups

            flops_numpy = (
                2 * np.prod(kernel_size) * num_in_channels * filters_per_channel * np.prod(output_shapes[name])
            )
            weights_numpy = np.prod(kernel_size) * num_in_channels * filters_per_channel

            flops[name] = flops_numpy.astype(int).item()
            weights[name] = weights_numpy.astype(int).item()

        for node in graph.get_nodes_by_metatypes(self._linear_op_metatypes):
            name = node.node_name
            if name in op_addresses_to_skip:
                continue

            num_in_features = input_channels.get(name, node.layer_attributes.in_features)
            num_out_features = output_channels.get(name, node.layer_attributes.out_features)

            flops_numpy = 2 * num_in_features * num_out_features * np.prod(output_shapes[name][:-1])
            weights_numpy = num_in_features * num_out_features
            flops[name] = flops_numpy
            weights[name] = weights_numpy

        return flops, weights

    def count_filters_num(self, graph: NNCFGraph, output_channels: Dict[NNCFNodeName, int] = None) -> int:
        """
        Counts filters of `op_metatypes` layers taking into account new output channels number.

        :param graph: NNCFGraph.
        :param output_channels:  A dictionary of output channels number in pruned model.
        :return: Current number of filters according to given graph and output channels.
        """
        filters_num = 0
        output_channels = output_channels or {}
        for node in graph.get_nodes_by_metatypes(self._conv_op_metatypes + self._linear_op_metatypes):
            filters_num += output_channels.get(node.node_name, get_output_channels(node))
        return filters_num



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from nncf.common.graph.patterns import GraphPattern
from nncf.common.graph.patterns import Patterns
from nncf.tensorflow.graph.metatypes.tf_ops import TFIdentityOpMetatype
from nncf.tensorflow.graph.pattern_operations import ATOMIC_ACTIVATIONS_OPERATIONS
from nncf.tensorflow.graph.pattern_operations import BATCH_NORMALIZATION_OPERATIONS
from nncf.tensorflow.graph.pattern_operations import ELEMENTWISE_OPERATIONS
from nncf.tensorflow.graph.pattern_operations import LINEAR_OPERATIONS
from nncf.tensorflow.graph.pattern_operations import QUANTIZATION_AGNOSTIC_OPERATIONS
from nncf.tensorflow.graph.patterns import create_conv2d_biasadd_pattern
from nncf.tensorflow.graph.patterns import create_h_sigmoid_act
from nncf.tensorflow.graph.patterns import create_h_swish_act
from nncf.tensorflow.graph.patterns import create_matmul_biasadd_pattern


def _get_tf_hw_fused_patterns() -> Patterns:
    retval = Patterns()
    linear_ops = GraphPattern()
    linear_ops.add_node(**LINEAR_OPERATIONS)

    eltwise_ops = GraphPattern()
    eltwise_ops.add_node(**ELEMENTWISE_OPERATIONS)

    batch_norm = GraphPattern()
    batch_norm.add_node(**BATCH_NORMALIZATION_OPERATIONS)

    h_sigmoid = create_h_sigmoid_act()
    h_swish = create_h_swish_act()
    retval.register(h_sigmoid, "H_SIGMOID", match=True)
    retval.register(h_swish, "H_SWISH", match=True)

    matmul_biasadd = create_matmul_biasadd_pattern()
    retval.register(matmul_biasadd, "MATMUL_BIASADD", match=True)

    conv2d_biasadd = create_conv2d_biasadd_pattern()
    retval.register(conv2d_biasadd, "CONV2D_BIASADD", match=True)

    atomic_activations = GraphPattern()
    atomic_activations.add_node(**ATOMIC_ACTIVATIONS_OPERATIONS)
    activations = atomic_activations | h_swish | h_sigmoid
    batch_norm_activations_permutation = batch_norm + activations | activations + batch_norm
    any_bn_act_combo = batch_norm | activations | batch_norm_activations_permutation

    identity = GraphPattern()
    identity.add_node(type=TFIdentityOpMetatype.get_all_aliases(), label="IDENTITY")
    linear_ops_maybe_followed_by_identity = linear_ops | (linear_ops + identity) | matmul_biasadd | conv2d_biasadd

    agnostic_ops = GraphPattern()
    agnostic_ops.add_node(**QUANTIZATION_AGNOSTIC_OPERATIONS)
    any_ag_bn_act_combo = agnostic_ops + activations | any_bn_act_combo

    retval.register(linear_ops_maybe_followed_by_identity, name="LINEAR", match=True)
    retval.register(batch_norm_activations_permutation, name="BN_ACT_OR_ACT_BN", match=True)
    retval.register(
        linear_ops_maybe_followed_by_identity + any_ag_bn_act_combo, "LINEAR + ANY_AG_BN_ACT_COMBO", match=True
    )
    retval.register(eltwise_ops + any_ag_bn_act_combo, "ELTWISE + ANY_AG_BN_ACT_COMBO", match=True)
    return retval


TF_HW_FUSED_PATTERNS = _get_tf_hw_fused_patterns()



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Callable, Dict, List, Optional, Union

from nncf.common.graph.transformations.commands import TargetPoint
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.graph.transformations.commands import TransformationCommand
from nncf.common.graph.transformations.commands import TransformationPriority
from nncf.common.graph.transformations.commands import TransformationType
from nncf.common.stateful_classes_registry import TF_STATEFUL_CLASSES


class TFLayerPointStateNames:
    LAYER_NAME = "layer_name"
    TARGET_TYPE = "target_type"


@TF_STATEFUL_CLASSES.register()
class TFLayerPoint(TargetPoint):
    """
    `TFLayerPoint` defines an object or spot relative to the layer in the
    TensorFlow model graph. It can be the layer itself, layer weights, specific
    spots in the model graph, for example, insertion spots before/after layer
    and etc.
    """

    _state_names = TFLayerPointStateNames

    def __init__(self, target_type: TargetType, layer_name: str):
        super().__init__(target_type)
        self._layer_name = layer_name

    @property
    def layer_name(self) -> str:
        return self._layer_name

    def __eq__(self, other: "TFLayerPoint") -> bool:
        if isinstance(other, TFLayerPoint):
            return self.type == other.type and self.layer_name == other.layer_name
        return False

    def __str__(self) -> str:
        return super().__str__() + " " + self.layer_name

    def get_state(self) -> Dict[str, Any]:
        """
        Returns a dictionary with Python data structures (dict, list, tuple, str, int, float, True, False, None) that
        represents state of the object.

        :return: state of the object
        """
        return {
            self._state_names.TARGET_TYPE: self._target_type.get_state(),
            self._state_names.LAYER_NAME: self.layer_name,
        }

    @classmethod
    def from_state(cls, state: Dict[str, Any]) -> "TFLayerPoint":
        """
        Creates the object from its state.

        :param state: Output of `get_state()` method.
        """
        kwargs = {
            cls._state_names.TARGET_TYPE: TargetType.from_state(state[cls._state_names.TARGET_TYPE]),
            cls._state_names.LAYER_NAME: state[cls._state_names.LAYER_NAME],
        }
        return cls(**kwargs)


class TFMultiLayerPoint:
    """
    `TFMultiLayerPoint` stores a list of target points that will be
    combined into shared callable object. Each point can be spots in the model
    graph: insertion spots before/after layer.
    """

    def __init__(self, target_points: List[TargetPoint]):
        self._target_points = target_points

    @property
    def target_points(self) -> List[TargetPoint]:
        return self._target_points

    def __str__(self) -> str:
        return f"TFMultiLayerPoint: {[str(t) for t in self._target_points]}"


class TFLayerStateNames:
    LAYER_NAME = "layer_name"


@TF_STATEFUL_CLASSES.register()
class TFLayer(TFLayerPoint):
    """
    `TFLayer` defines a layer in the TensorFlow model graph.

    For example, `TFLayer` is used to specify the layer in the removal command
    to remove from the model.
    """

    _state_names = TFLayerStateNames

    def __init__(self, layer_name: str):
        super().__init__(TargetType.LAYER, layer_name)

    def get_state(self) -> Dict[str, Any]:
        """
        Returns a dictionary with Python data structures (dict, list, tuple, str, int, float, True, False, None) that
        represents state of the object.

        :return: state of the object
        """
        return {
            self._state_names.LAYER_NAME: self.layer_name,
        }

    @classmethod
    def from_state(cls, state: Dict[str, Any]) -> "TFLayer":
        """
        Creates the object from its state.

        :param state: Output of `get_state()` method.
        """
        return cls(**state)


class TFBeforeLayerStateNames:
    LAYER_NAME = "layer_name"
    INSTANCE_IDX = "instance_idx"
    INPUT_PORT_ID = "input_port_id"


@TF_STATEFUL_CLASSES.register()
class TFBeforeLayer(TFLayerPoint):
    """
    `TFBeforeLayer` defines a spot before the layer in the TensorFlow model graph.

    For example, `TFBeforeLayer` is used in the insertion commands to specify
    where the new object should be inserted.
    """

    _state_names = TFBeforeLayerStateNames

    def __init__(self, layer_name: str, instance_idx: int = 0, input_port_id: int = 0):
        super().__init__(TargetType.BEFORE_LAYER, layer_name)
        self._instance_idx = instance_idx
        self._input_port_id = input_port_id

    @property
    def instance_idx(self) -> int:
        return self._instance_idx

    @property
    def input_port_id(self) -> int:
        return self._input_port_id

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, TFBeforeLayer)
            and self.layer_name == other.layer_name
            and self.instance_idx == other.instance_idx
            and self.input_port_id == other.input_port_id
        )

    def __str__(self) -> str:
        return " ".join([super().__str__(), str(self.instance_idx), str(self.input_port_id)])

    def __hash__(self) -> int:
        return hash(str(self))

    def get_state(self) -> Dict[str, Any]:
        """
        Returns a dictionary with Python data structures (dict, list, tuple, str, int, float, True, False, None) that
        represents state of the object.

        :return: state of the object
        """
        return {
            self._state_names.LAYER_NAME: self.layer_name,
            self._state_names.INSTANCE_IDX: self.instance_idx,
            self._state_names.INPUT_PORT_ID: self.input_port_id,
        }

    @classmethod
    def from_state(cls, state: Dict[str, Any]) -> "TFBeforeLayer":
        """
        Creates the object from its state.

        :param state: Output of `get_state()` method.
        """
        return cls(**state)


class TFAfterLayerStateNames:
    LAYER_NAME = "layer_name"
    INSTANCE_IDX = "instance_idx"
    OUTPUT_PORT_ID = "output_port_id"


@TF_STATEFUL_CLASSES.register()
class TFAfterLayer(TFLayerPoint):
    """
    `TFAfterLayer` defines a spot after the layer in the TensorFlow model graph.

    For example, `TFAfterLayer` is used in the insertion commands to specify
    where the new object should be inserted.
    """

    _state_names = TFAfterLayerStateNames

    def __init__(self, layer_name: str, instance_idx: int = 0, output_port_id: int = 0):
        super().__init__(TargetType.AFTER_LAYER, layer_name)
        self._instance_idx = instance_idx
        self._output_port_id = output_port_id

    @property
    def instance_idx(self) -> int:
        return self._instance_idx

    @property
    def output_port_id(self) -> int:
        return self._output_port_id

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, TFAfterLayer)
            and self.type == other.type
            and self.layer_name == other.layer_name
            and self.instance_idx == other.instance_idx
            and self._output_port_id == other.output_port_id
        )

    def __str__(self) -> str:
        return " ".join([super().__str__(), str(self.instance_idx), str(self.output_port_id)])

    def __hash__(self) -> int:
        return hash(str(self))

    def get_state(self) -> Dict[str, Any]:
        """
        Returns a dictionary with Python data structures (dict, list, tuple, str, int, float, True, False, None) that
        represents state of the object.

        :return: state of the object
        """
        return {
            self._state_names.LAYER_NAME: self.layer_name,
            self._state_names.INSTANCE_IDX: self.instance_idx,
            self._state_names.OUTPUT_PORT_ID: self.output_port_id,
        }

    @classmethod
    def from_state(cls, state: Dict[str, Any]) -> "TFAfterLayer":
        """
        Creates the object from its state.

        :param state: Output of `get_state()` method.
        """
        return cls(**state)


class TFLayerWeightsStateNames:
    LAYER_NAME = "layer_name"
    WEIGHTS_ATTR_NAME = "weights_attr_name"


@TF_STATEFUL_CLASSES.register()
class TFLayerWeight(TFLayerPoint):
    """
    `TFLayerWeight` defines the layer weights.

    For example, `TFLayerWeight` is used in the insertion command to specify
    the layer weights for which an operation with weights should be inserted.
    """

    _state_names = TFLayerWeightsStateNames

    def __init__(self, layer_name: str, weights_attr_name: str):
        super().__init__(TargetType.OPERATION_WITH_WEIGHTS, layer_name)
        self._weights_attr_name = weights_attr_name

    @property
    def weights_attr_name(self) -> str:
        return self._weights_attr_name

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, TFLayerWeight)
            and self.type == other.type
            and self.layer_name == other.layer_name
            and self.weights_attr_name == other.weights_attr_name
        )

    def __str__(self) -> str:
        return super().__str__() + " " + self.weights_attr_name

    def __hash__(self) -> int:
        return hash(str(self))

    def get_state(self) -> Dict[str, Any]:
        """
        Returns a dictionary with Python data structures (dict, list, tuple, str, int, float, True, False, None) that
        represents state of the object.

        :return: state of the object
        """
        return {
            self._state_names.LAYER_NAME: self.layer_name,
            self._state_names.WEIGHTS_ATTR_NAME: self.weights_attr_name,
        }

    @classmethod
    def from_state(cls, state: Dict[str, Any]) -> "TFLayerWeight":
        """
        Creates the object from its state.

        :param state: Output of `get_state()` method.
        """
        return cls(**state)


class TFOperationWithWeightsStateNames:
    LAYER_NAME = "layer_name"
    WEIGHTS_ATTR_NAME = "weights_attr_name"
    OPERATION_NAME = "operation_name"


@TF_STATEFUL_CLASSES.register()
class TFOperationWithWeights(TFLayerWeight):
    """
    `TFOperationWithWeights` defines an operation with weights.

    For example, `TFOperationWithWeights` is used to specify the operation with
    weights in the removal command to remove from the model.
    """

    _state_names = TFOperationWithWeightsStateNames

    def __init__(self, layer_name: str, weights_attr_name: str, operation_name: str):
        super().__init__(layer_name, weights_attr_name)
        self._operation_name = operation_name

    @property
    def operation_name(self) -> str:
        return self._operation_name

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, TFOperationWithWeights)
            and self.type == other.type
            and self.layer_name == other.layer_name
            and self.weights_attr_name == other.weights_attr_name
            and self.operation_name == other.operation_name
        )

    def __str__(self) -> str:
        return super().__str__() + " " + self.operation_name

    def __hash__(self) -> int:
        return hash(str(self))

    def get_state(self) -> Dict[str, Any]:
        """
        Returns a dictionary with Python data structures (dict, list, tuple, str, int, float, True, False, None) that
        represents state of the object.

        :return: state of the object
        """
        return {
            self._state_names.LAYER_NAME: self._layer_name,
            self._state_names.WEIGHTS_ATTR_NAME: self._weights_attr_name,
            self._state_names.OPERATION_NAME: self._operation_name,
        }

    @classmethod
    def from_state(cls, state: Dict[str, Any]) -> "TFOperationWithWeights":
        """
        Creates the object from its state.

        :param state: Output of `get_state()` method.
        """
        return cls(**state)


class TFTransformationCommand(TransformationCommand):
    """
    The base class for all Tensorflow transformation commands.
    """

    def __init__(self, command_type: TransformationType, target_point: TargetPoint):
        """
        Constructor.

        :param command_type: Type of the transformation command.
        :param target_point: Target point, the object or spot in the model graph
            to which the transformation command will be applied.
        """
        super().__init__(command_type, target_point)

    def check_command_compatibility(self, command: "TFTransformationCommand") -> bool:
        return (
            isinstance(command, TFTransformationCommand)
            and self.type == command.type
            and self.target_point == command.target_point
        )

    def union(self, other: "TFTransformationCommand") -> "TFTransformationCommand":
        raise NotImplementedError()

    def __add__(self, other: "TFTransformationCommand") -> "TFTransformationCommand":
        return self.union(other)


class TFInsertionCommand(TFTransformationCommand):
    """
    Inserts objects at the target point in the TensorFlow model graph.
    """

    def __init__(
        self,
        target_point: Union[TargetPoint, TFMultiLayerPoint],
        callable_object: Optional[Callable] = None,
        priority: Optional[TransformationPriority] = None,
    ):
        super().__init__(TransformationType.INSERT, target_point)
        self.callable_objects = []
        if callable_object is not None:
            _priority = TransformationPriority.DEFAULT_PRIORITY if priority is None else priority
            self.callable_objects.append((callable_object, _priority))

    @property
    def insertion_objects(self) -> List[Callable]:
        return [x for x, _ in self.callable_objects]

    def union(self, other: TFTransformationCommand) -> "TFInsertionCommand":
        if isinstance(self.target_point, TFMultiLayerPoint):
            raise NotImplementedError(
                "A command of TFInsertionCommand type with TFMultiLayerPoint "
                "could not be united with another command"
            )

        if not self.check_command_compatibility(other):
            raise ValueError("{} and {} commands could not be united".format(type(self).__name__, type(other).__name__))

        com = TFInsertionCommand(self.target_point)
        com.callable_objects = self.callable_objects + other.callable_objects
        com.callable_objects = sorted(com.callable_objects, key=lambda x: x[1])
        return com


class TFRemovalCommand(TFTransformationCommand):
    """
    Removes the target object.
    """

    def __init__(self, target_point: TargetPoint):
        super().__init__(TransformationType.REMOVE, target_point)

    def union(self, other: TFTransformationCommand) -> "TFRemovalCommand":
        raise NotImplementedError("A command of TFRemovalCommand type could not be united with another command")


class TFMultipleInsertionCommands(TFTransformationCommand):
    """
    A list of insertion commands combined by a common global target point but
    with different target points in between.

    For example, If a layer has multiple weight variables you can use this
    transformation command to insert operations with weights for each layer
    weights variable at one multiple insertion command.
    """

    def __init__(
        self,
        target_point: TargetPoint,
        check_target_points_fn: Optional[Callable] = None,
        commands: Optional[List[TFTransformationCommand]] = None,
    ):
        super().__init__(TransformationType.MULTI_INSERT, target_point)
        self.check_target_points_fn = check_target_points_fn
        if check_target_points_fn is None:
            self.check_target_points_fn = lambda tp0, tp1: tp0 == tp1
        self._commands = []
        if commands is not None:
            for cmd in commands:
                self.add_insertion_command(cmd)

    @property
    def commands(self) -> List[TFTransformationCommand]:
        return self._commands

    def check_insertion_command(self, command: TFTransformationCommand) -> bool:
        if (
            isinstance(command, TFTransformationCommand)
            and command.type == TransformationType.INSERT
            and self.check_target_points_fn(self.target_point, command.target_point)
        ):
            return True
        return False

    def add_insertion_command(self, command: TFTransformationCommand) -> None:
        if not self.check_insertion_command(command):
            raise ValueError("{} command could not be added".format(type(command).__name__))

        for idx, cmd in enumerate(self.commands):
            if cmd.target_point == command.target_point:
                self.commands[idx] = cmd + command
                break
        else:
            self.commands.append(command)

    def union(self, other: TFTransformationCommand) -> "TFMultipleInsertionCommands":
        if not self.check_command_compatibility(other):
            raise ValueError("{} and {} commands could not be united".format(type(self).__name__, type(other).__name__))

        def make_check_target_points_fn(fn1, fn2):
            def check_target_points(tp0, tp1):
                return fn1(tp0, tp1) or fn2(tp0, tp1)

            return check_target_points

        check_target_points_fn = (
            self.check_target_points_fn
            if self.check_target_points_fn == other.check_target_points_fn
            else make_check_target_points_fn(self.check_target_points_fn, other.check_target_points_fn)
        )

        multi_cmd = TFMultipleInsertionCommands(self.target_point, check_target_points_fn, self.commands)
        for cmd in other.commands:
            multi_cmd.add_insertion_command(cmd)
        return multi_cmd
